# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                  initial documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable diagonal model
=======================

This module implements the concept of scalable diagonal model,
which is a particular scalable model built from an input-output
dataset relying on a diagonal design of experiments (DOE)
where inputs vary proportionally from their lower bounds
to their upper bounds, following the diagonal of the input space.

So for every output, the dataset catches its evolution
with respect to this proportion, which makes it a monodimensional behavior.
Then, for a new user-defined problem dimension,
the scalable model extrapolates this monodimensional behavior
to the different input directions.

The concept of scalable diagonal model is implemented through
the :class:`.ScalableDiagonalModel` class
which is composed of a :class:`.ScalableDiagonalApproximation`.
With regard to the diagonal DOE, |g| proposes the
:class:`.DiagonalDOE` class.
"""
from __future__ import division, unicode_literals

import logging
from numbers import Number

import matplotlib.pyplot as plt
from numpy import arange, argsort, array, array_equal, atleast_1d, hstack
from numpy import max as np_max
from numpy import mean, median
from numpy import min as np_min
from numpy import nan_to_num, sqrt, vstack, where, zeros
from numpy.random import choice, rand, randint
from numpy.random import seed as npseed
from past.utils import old_div
from scipy.interpolate import InterpolatedUnivariateSpline

from gemseo.problems.scalable.data_driven.model import ScalableModel
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.matplotlib_figure import save_show_figure
from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)


class ScalableDiagonalModel(ScalableModel):
    """Scalable diagonal model."""

    ABBR = "sdm"

    def __init__(
        self,
        data,
        sizes=None,
        fill_factor=-1,
        comp_dep=None,
        inpt_dep=None,
        force_input_dependency=False,
        allow_unused_inputs=True,
        seed=1,
        group_dep=None,
    ):
        """Constructor.

        :param Dataset data: learning dataset.
        :param dict sizes: sizes of input and output variables.
            If None, use the original sizes.
            Default: None.
        :param fill_factor: degree of sparsity of the dependency matrix.
            Default: -1.
        :param comp_dep: matrix that establishes the selection
            of a single original component for each scalable component
        :param inpt_dep: dependency matrix that establishes the
            dependency of outputs wrt inputs
        :param bool force_input_dependency: for any output, force dependency
            with at least on input.
        :param bool allow_unused_inputs: possibility to have an input
            with no dependence with any output
        :param int seed: seed
        :param dict(list(str)) group_dep: dependency
            between inputs and outputs
        """
        if isinstance(fill_factor, Number):
            fill_factor = {
                function_name: fill_factor
                for function_name in data.get_names(data.OUTPUT_GROUP)
            }
        elif not isinstance(fill_factor, dict):
            raise TypeError(
                "Fill factor must be either a number between 0 and 1, "
                "a number equal to -1 or a dictionary."
            )
        parameters = {
            "fill_factor": fill_factor,
            "comp_dep": comp_dep,
            "inpt_dep": inpt_dep,
            "force_input_dependency": force_input_dependency,
            "allow_unused_inputs": allow_unused_inputs,
            "seed": seed,
            "group_dep": group_dep,
        }
        super(ScalableDiagonalModel, self).__init__(data, sizes, **parameters)
        self.t_scaled, self.f_scaled = self.__build_scalable_functions()

    def __build_dependencies(self):
        """Build dependencies.

        :return: matrix that establishes the selection of a single original
            component for each scalable component, dependency matrix that
            establishes the dependency of outputs wrt inputs.
        :rtype: ndarray, ndarray
        """
        comp_dep = self.parameters["comp_dep"]
        inpt_dep = self.parameters["inpt_dep"]
        if comp_dep is None or inpt_dep is None:
            comp_dep, inpt_dep = self.generate_random_dependency()
        return comp_dep, inpt_dep

    def scalable_function(self, input_value=None):
        """Evaluate the scalable functions.

        :param dict input_value: input values.
            If None, use default inputs.
        :return: evaluation of the scalable functions.
        :rtype: dict
        """
        dict_to_array = DataConversion.dict_to_array
        input_value = input_value or self.default_inputs
        input_value = dict_to_array(input_value, self.inputs_names)
        scal_func = self.model.get_scalable_function
        return {fname: scal_func(fname)(input_value) for fname in self.outputs_names}

    def scalable_derivatives(self, input_value=None):
        """Evaluate the scalable derivatives.

        :param dict input_value: input values.
            If None, use default inputs.
        :return: evaluation of the scalable derivatives.
        :rtype: dict
        """
        dict_to_array = DataConversion.dict_to_array
        input_value = input_value or self.default_inputs
        input_value = dict_to_array(input_value, self.inputs_names)
        scal_der = self.model.get_scalable_derivative
        return {fname: scal_der(fname)(input_value) for fname in self.outputs_names}

    def build_model(self):
        """Build model with original sizes for input and output variables.

        :return: scalable approximation.
        :rtype: ScalableDiagonalApproximation
        """
        comp_dep, inpt_dep = self.__build_dependencies()
        seed = self.parameters["seed"]
        scalable_approximation = ScalableDiagonalApproximation(
            self.sizes, comp_dep, inpt_dep, seed
        )
        return scalable_approximation

    def __build_scalable_functions(self):
        """Builds all the required functions from the original dataset."""
        t_scaled = {}
        f_scaled = {}
        for function_name in self.outputs_names:
            t_sc, f_sc = self.model.build_scalable_function(
                function_name, self.data, self.inputs_names
            )
            t_scaled[function_name] = t_sc
            f_scaled[function_name] = f_sc
        return t_scaled, f_scaled

    def __get_variables_locations(self, names):
        """Get the locations of first component of each variables.

        :param names: list of variables names.
        :type names: list(str)
        :return: list of locations.
        :rtype: list(int)
        """
        positions = []
        current_position = 0
        for name in names:
            positions.append(current_position)
            current_position += self.sizes[name]
        return positions

    def __convert_dependency_to_array(self, dependency):
        """Convert a dependency object of type dictionary into a dependency object of
        type array.

        :param dependency: input-output dependency structure.
        :type dependency: dict
        :return: dependency matrix.
        :rtype: array
        """
        matrix = None
        for output_name in self.outputs_names:
            row = hstack([dependency[output_name][inpt] for inpt in self.inputs_names])
            matrix = row if matrix is None else vstack((matrix, row))
        return matrix.T

    def plot_dependency(
        self, add_levels=True, save=True, show=False, directory=".", png=False
    ):
        """This method plots the dependency matrix of a discipline in the form of a
        chessboard, where rows represent inputs, columns represent output and gray scale
        represent the dependency level between inputs and outputs.

        :param bool add_levels: add values of dependency levels in percentage.
            Default: True.
        :param bool save: if True, export the plot into a file.
            Default: True.
        :param bool show: if True, display the plot.
            Default: False.
        :param str directory: directory path. Default: '.'.
        :param bool png: if True, the file format is PNG. Otherwise, use PDF.
            Default: False.
        """
        inputs_positions = self.__get_variables_locations(self.inputs_names)
        outputs_positions = self.__get_variables_locations(self.outputs_names)
        dependency = self.model.io_dependency
        dependency_matrix = self.__convert_dependency_to_array(dependency)
        is_binary_matrix = array_equal(
            dependency_matrix, dependency_matrix.astype(bool)
        )
        if not is_binary_matrix:
            dp_sum = dependency_matrix.sum(0)
            dependency_matrix = old_div(dependency_matrix, dp_sum)

        fig, axes = plt.subplots()
        axes.matshow(dependency_matrix, cmap="Greys", vmin=0)
        axes.set_yticks(inputs_positions)
        axes.set_yticklabels(self.inputs_names)
        axes.set_xticks(outputs_positions)
        axes.set_xticklabels(self.outputs_names, ha="left", rotation="vertical")
        axes.tick_params(axis="both", which="both", length=0)
        for pos in inputs_positions[1:]:
            axes.axhline(y=pos - 0.5, color="r", linewidth=0.5)
        for pos in outputs_positions[1:]:
            axes.axvline(x=pos - 0.5, color="r", linewidth=0.5)
        if add_levels and not is_binary_matrix:
            for i in range(dependency_matrix.shape[0]):
                for j in range(dependency_matrix.shape[1]):
                    val = int(round(dependency_matrix[i, j] * 100))
                    med = median(dependency_matrix[:, j] * 100)
                    col = "white" if val > med else "black"
                    axes.text(j, i, val, ha="center", va="center", color=col)
        if save:
            extension = "png" if png else "pdf"
            file_path = Path(directory) / "{}_dependency.{}".format(
                self.name, extension
            )
        else:
            file_path = None
        save_show_figure(fig, show, file_path)
        return str(file_path)

    def plot_1d_interpolations(
        self, save=False, show=False, step=0.01, varnames=None, directory=".", png=False
    ):
        r"""This methods plots the scaled 1D interpolations,
        a.k.a. basis functions.

        A basis function is a monodimensional function
        interpolating the samples of a given output component
        over the input sampling line
        :math:`t\in[0,1]\mapsto \\underline{x}+t(\overline{x}-\\underline{x})`.

        There are as many basis functions
        as there are output components from the discipline.
        Thus, for a discipline with a single output in dimension 1,
        there is 1 basis function.
        For a discipline with a single output in dimension 2,
        there are 2 basis functions.
        For a discipline with an output in dimension 2
        and an output in dimension 13,
        there are 15 basis functions. And so on.
        This method allows to plot the basis functions associated
        with all outputs or only part of them,
        either on screen (:code:`show=True`), in a file (:code:`save=True`)
        or both.
        We can also specify the discretization :code:`step`
        whose default value is :code:`0.01`.

        :param bool save: if True, export the plot as a PDF file
            (Default value = False)
        :param bool show: if True, display the plot (Default value = False)
        :param bool step: Step to evaluate the 1d interpolation function
            (Default value = 0.01)
        :param list(str) varnames: names of the variable to plot;
            if None, all variables are plotted (Default value = None)
        :param str directory: directory path. Default: '.'.
        :param bool png: if True, the file format is PNG. Otherwise, use PDF.
            Default: False.
        """
        function_names = varnames or self.outputs_names
        x_vals = arange(0.0, 1.0 + step, step)
        fnames = []
        for func in function_names:
            components = self.model.interpolation_dict[func]
            for index, function in enumerate(components):
                plt.figure()
                doe_t = self.t_scaled[func]
                doe_f = [output[index] for output in self.f_scaled[func]]
                y_vals = function(x_vals)
                plt.xlim(-0.05, 1.05)
                plt.ylim(-0.1, 1.1)
                plt.plot(
                    x_vals,
                    old_div((y_vals - min(y_vals)), (max(y_vals) - min(y_vals))),
                    label=func + str(index),
                )
                plt.plot(
                    doe_t,
                    old_div((doe_f - min(y_vals)), (max(y_vals) - min(y_vals))),
                    "or",
                )
                plt.xlabel("Scaled abscissa", fontsize=18)
                plt.ylabel("Interpolation value", fontsize=18)
                plt.title("1D interpolation of " + self.name + "." + func)
                fig = plt.gcf()
                if save:
                    extension = "png" if png else "pdf"
                    file_path = Path(directory) / "{}_{}_1D_interpolation_{}.{}".format(
                        self.name, func, index, extension
                    )
                    fnames.append(str(file_path))
                else:
                    file_path = None
                save_show_figure(fig, show, file_path)
        return fnames

    def generate_random_dependency(self):
        """Generates a random dependency structure for use in scalable discipline.

        :return: output component dependency and input-output dependency
        :rtype: dict(int), dict(dict(array))
        """
        npseed(self.parameters["seed"])
        io_dependency = self.parameters["group_dep"] or {}
        for function_name in self.outputs_names:
            input_names = io_dependency.get(function_name, self.inputs_names)
            io_dependency[function_name] = input_names

        if self.parameters.get("inpt_dep") is None:
            io_dep = self.__generate_random_io_dep(io_dependency)

        if self.parameters.get("comp_dep") is None:
            out_map = self.__generate_random_output_map()

        # If an output function does not have any dependency with inputs,
        # add a random dependency if independent outputs are forbidden
        if self.parameters["force_input_dependency"]:
            for function_name in self.outputs_names:
                for function_component in range(self.sizes.get(function_name)):
                    self.__complete_random_dep(
                        io_dep, function_name, function_component, io_dependency
                    )

        # If an input parameter does not have any dependency with output
        # functions, add a random dependency if unused inputs are not allowed
        if not self.parameters["allow_unused_inputs"]:
            for input_name in self.inputs_names:
                for input_component in range(self.sizes.get(input_name)):
                    self.__complete_random_dep(
                        io_dep, input_name, input_component, io_dependency
                    )
        return out_map, io_dep

    def __generate_random_io_dep(self, io_dependency):
        """Generate the dependency between the new inputs and the new outputs.

        :param io_dependency: input-output dependency structure. If None,
            all output components can depend on all input components.
            Default: None.
        :type io_dependency: dict(list(str))
        :return: random input-output dependencies
        :rtype: dict(dict(array))
        """
        error_msg = (
            "Fill factor must be a number, "
            "either -1 or a real number between 0 and 1."
        )
        r_io_dependency = {}
        for function_name in self.outputs_names:
            fill_factor = self.parameters["fill_factor"].get(function_name, 0.7)
            if fill_factor != -1.0 and (fill_factor < 0.0 or fill_factor > 1):
                raise TypeError(error_msg)
            function_size = self.sizes.get(function_name)
            r_io_dependency[function_name] = {}
            for input_name in self.inputs_names:
                input_size = self.sizes.get(input_name)
                if input_name in io_dependency[function_name]:
                    if 0 <= fill_factor <= 1:
                        rand_dep = choice(
                            2,
                            (function_size, input_size),
                            p=[1.0 - fill_factor, fill_factor],
                        )
                    else:
                        rand_dep = rand(function_size, input_size)
                    r_io_dependency[function_name][input_name] = rand_dep
                else:
                    zeros_dep = zeros((function_size, input_size))
                    r_io_dependency[function_name][input_name] = zeros_dep
        return r_io_dependency

    def __generate_random_output_map(self):
        """Generate the dependency between the original and new output components for
        the different outputs.

        :return: component dependencies
        :rtype: dict(int)
        """
        out_map = {}
        for function_name in self.outputs_names:
            original_function_size = self.original_sizes.get(function_name)
            function_size = self.sizes.get(function_name)
            out_map[function_name] = randint(original_function_size, size=function_size)
        return out_map

    def __complete_random_dep(self, r_io_dep, dataname, index, io_dep):
        """Complete random dependency if row (input name) or column (function name) of
        the random dependency matrix is empty.

        :param array r_io_dep: input-output dependency.
        :param str dataname: name of the variable to check
            if component is empty.
        :param int index: component index of the variable.
        :param io_dep: input-output dependency structure. If None,
            all output components can depend on all input components.
            Default: None.
        :type io_dep: dict(list(str))
        """
        is_input = dataname in self.inputs_names
        if is_input:
            varnames = []
            for function_name, inputs in io_dep.items():
                if dataname in inputs:
                    varnames.append(function_name)
            inpt_dep_mat = hstack(
                [r_io_dep[varname][dataname].T for varname in varnames]
            )
        else:
            varnames = io_dep[dataname]
            inpt_dep_mat = hstack([r_io_dep[dataname][varname] for varname in varnames])
        if sum(inpt_dep_mat[index, :]) == 0:
            prob = [self.sizes.get(varname, 1) for varname in varnames]
            prob = [float(x) / sum(prob) for x in prob]
            id_var = choice(len(varnames), p=prob)
            id_comp = randint(0, self.sizes.get(varnames[id_var], 1))
            if is_input:
                varname = varnames[id_var]
                r_io_dep[varname][dataname][id_comp, index] = 1
            else:
                varname = varnames[id_var]
                r_io_dep[dataname][varname][index, id_comp] = 1


class ScalableDiagonalApproximation(object):
    """Methodology that captures the trends of a physical problem, and extends it into a
    problem that has scalable input and outputs dimensions The original and the
    resulting scalable problem have the same interface:

    all inputs and outputs have the same names; only their dimensions vary.
    """

    def __init__(self, sizes, output_dependency, io_dependency, seed=0):
        """
        Constructor:

        :param sizes: sizes of both input and output variables.
        :type sizes: dict
        :param output_dependency: dependency between old and new outputs.
        :type output_dependency: dict
        :param io_dependency: dependency between new inputs and new outputs.
        :type io_dependency: dict
        """
        super(ScalableDiagonalApproximation, self).__init__()
        self.sizes = sizes
        # dependency matrices
        self.output_dependency = output_dependency
        self.io_dependency = io_dependency
        # dictionaries of interpolations and extrapolations
        self.interpolation_dict = {}
        self.d_interpolation_dict = {}
        self.interpolators_dict = {}
        self.scalable_functions = {}
        self.scalable_dfunctions = {}
        # seed for random generator
        npseed(seed)

    def build_scalable_function(self, function_name, dataset, input_names, degree=3):
        """Build interpolation interpolation from a 1D input and output function. Add
        the model to the local dictionary.

        :param str function_name: name of the output function
        :param Dataset dataset: the input-output dataset
        :param list(str) input_names: names of the input variables
        :param int degree: degree of interpolation (Default value = 3)
        """
        x2_scaled = nan_to_num(dataset.get_data_by_group(dataset.INPUT_GROUP) ** 2)
        t_scaled = [atleast_1d(sqrt(mean(val.real))) for val in x2_scaled]
        f_scaled = dataset[function_name][function_name].real

        # sort t_scaled and f_scaled following the t_scaled ascending order
        indices = argsort([val[0] for val in t_scaled])
        t_scaled = [t_scaled[index] for index in indices]
        f_scaled = [f_scaled[index] for index in indices]

        # scale the output samples: [a1, b1] x ... x [am, bm] -> [0, 1]^m
        f_scaled = self.scale_samples(f_scaled)

        # interpolate the (t_scaled, f_scaled) data
        self._interpolate(function_name, t_scaled, f_scaled, degree)

        # compute the total input and output sizes
        (input_size, output_size) = self._compute_sizes(function_name, input_names)

        # extrapolation
        self._extrapolate(function_name, input_names, input_size, output_size)

        return t_scaled, f_scaled

    def get_scalable_function(self, output_function):
        """Retrieve the scalable function generated from the original discipline.

        :param str output_function: name of the output function
        """
        return self.scalable_functions[output_function]

    def get_scalable_derivative(self, output_function):
        """Retrieve the (scalable) gradient of the scalable function generated from the
        original discipline.

        :param str output_function: name of the output function
        """
        return self.scalable_dfunctions[output_function]

    @staticmethod
    def scale_samples(samples):
        """Scale samples of array into [0, 1]

        :param samples: samples of multivariate array
        :type samples: list(array)
        :return: samples of multivariate array
        :rtype: array
        """
        samples = array(samples)
        col_min = np_min(samples, 0)
        col_max = np_max(samples, 0)
        scaled_samples = samples - col_min
        range_col = col_max - col_min
        scaling = where(abs(range_col) > 1e-6, range_col, 1)
        scaled_samples /= scaling
        return scaled_samples

    def _interpolate(self, function_name, t_scaled, f_scaled, degree=3):
        """Interpolate a set of samples (t, y(t)) with a polynomial spline.

        :param str function_name: name of the interpolated function
        :param list(list(float)) t_scaled: set of points
        :param list(list(float)) f_scaled: set of images
        :param int degree: degree of the polynomial interpolation
        """
        nb_components = f_scaled[0].size
        list_interpolations = []
        list_derivatives = []
        for component in range(nb_components):
            f_scaled_component = [output[component] for output in f_scaled]
            # compute interpolation
            interpolation = InterpolatedUnivariateSpline(
                t_scaled, f_scaled_component, k=degree
            )
            # store spline and derivative
            list_interpolations.append(interpolation)
            list_derivatives.append(interpolation.derivative())
        # store interpolation and derivatives
        self.interpolation_dict[function_name] = list_interpolations
        self.d_interpolation_dict[function_name] = list_derivatives

    def _compute_sizes(self, function_name, input_names):
        """Determine the size of the vector input and output.

        :param str function_name: function name
        :param list(str) input_names: input names
        """
        input_size = 0
        for input_name in input_names:
            input_size += self.sizes.get(input_name, 1)
        output_size = self.sizes.get(function_name, 1)  # default 1
        return input_size, output_size

    def _extrapolate(self, function_name, input_names, input_size, output_size):
        """Extrapolate a 1D function to arbitrary input and output dimensions. Generate
        a function that produces an output with a given size from an input with a given
        size, and its derivative.

        :param str function_name: name of the output function
        :param list(str) input_names: names of the inputs
        :param int input_size: size of the input vector
        :param int output_size: size of the output vector
        """
        # crop the matrices to the correct sizes and convert to array
        io_dependency = {
            input_name: dep_mat[0:output_size, 0 : self.sizes[input_name]]
            for (input_name, dep_mat) in self.io_dependency[function_name].items()
        }
        io_dependency = DataConversion.dict_to_array(io_dependency, input_names)

        # Convert the input-output dependency matrix to a list
        # where the i-th element is a list whose j-th element corresponds to
        # the degree of dependence between the i-th output component and the
        # i-th input.
        coefficients = [list(row) for row in io_dependency]

        # Get the 1D interpolation functions and their derivatives
        interpolated_fun_1d = self.interpolation_dict[function_name]
        interpolated_dfun_1d = self.d_interpolation_dict[function_name]

        # Get the indices of the 1D interpolation functions
        # associated with the components of the new output.
        components_list = self.output_dependency[function_name]

        def scalable_function(input_vars):
            """n-dimensional to size_output-dimensional extrapolated function.

            :param list(float) input_vars: vector of inputs
            :returns: size_output extrapolated output
            """
            result = zeros(output_size)
            for out_id in range(output_size):
                interp_id = components_list[out_id]
                coeffs = coefficients[out_id]
                tmp = [
                    coeffs[in_id] * interpolated_fun_1d[interp_id](in_val)
                    for in_id, in_val in enumerate(input_vars)
                ]
                tmp = old_div(sum(tmp), sum(coeffs))
                result[out_id] = tmp
            return result

        def scalable_derivative(input_vars):
            """n-dimensional to size_output-dimensional extrapolated function Jacobian.

            :param list(float) input_vars: vector of inputs
            :returns: size_output - sizeof input_vars extrapolated output
            """
            dresult = zeros([output_size, input_size])
            for out_id in range(output_size):
                interp_id = components_list[out_id]
                coeffs = coefficients[out_id]
                tmp = [
                    coeffs[in_id] * interpolated_dfun_1d[interp_id](in_val)
                    for in_id, in_val in enumerate(input_vars)
                ]
                tmp = old_div(array(tmp), sum(coeffs))
                dresult[out_id, :] = tmp
            return dresult

        self.scalable_functions[function_name] = scalable_function
        self.scalable_dfunctions[function_name] = scalable_derivative



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""This module contains a class to select a machine learning algorithm from a list.

Machine learning is used to find relations or underlying structures in data.
There is however no algorithm that is universally better than the others
for an arbitrary problem.
As for optimization, there is *no free lunch* for machine learning [wolpert]_.

.. [wolpert] Wolpert, David H.
   "The lack of a priori distinctions between learning algorithms."
   Neural computation 8.7 (1996): 1341-1390.

Provided a quality measure,
one can thus compare the performances of different machine learning algorithms.

This process can be easily performed using the class :class:`.MLAlgoSelection`.

A machine learning algorithm is built using a set of (hyper)parameters,
before the learning takes place.
In order to choose the best hyperparameters,
a simple grid search over different values may be sufficient.
The :class:`.MLAlgoSelection` does this.
It can also perform a more advanced form of optimization
than a simple grid search over predefined values,
using the class :class:`.MLAlgoCalibration`.

.. seealso::

   :mod:`~gemseo.mlearning.core.ml_algo`
   :mod:`~gemseo.mlearning.core.calibration`
"""
from __future__ import division, unicode_literals

from itertools import product
from typing import Optional, Sequence, Tuple, Union

from gemseo.algos.design_space import DesignSpace
from gemseo.core.dataset import Dataset
from gemseo.core.scenario import ScenarioInputDataType
from gemseo.mlearning.core.calibration import MLAlgoCalibration
from gemseo.mlearning.core.factory import MLAlgoFactory
from gemseo.mlearning.core.ml_algo import MLAlgo
from gemseo.mlearning.qual_measure.quality_measure import MLQualityMeasure
from gemseo.mlearning.qual_measure.quality_measure import (
    OptionType as MeasureOptionType,
)


class MLAlgoSelection(object):
    """Machine learning algorithm selector.

    Attributes:
        dataset (Dataset): The learning dataset.
        measure (str): The name of a quality measure
            to measure the quality of the machine learning algorithms.
        measure_options (Dict[str,Union[int,Dataset]]): The options for the method
            to evaluate the quality measure.
        factory (MLAlgoFactory): The factory
            used for the instantiation of machine learning algorithms.
        candidates (List[Tuple[MLAlgo,float]]): The candidate machine learning
            algorithms, after possible calibration, and their quality measures.
    """

    def __init__(
        self,
        dataset,  # type: Dataset
        measure,  # type: str
        eval_method=MLQualityMeasure.LEARN,  # type: str
        samples=None,  # type: Optional[Sequence[int]]
        **measure_options  # type:MeasureOptionType
    ):  # type: (...) -> None
        """
        Args:
            dataset: The learning dataset.
            measure: The name of a quality measure
                to measure the quality of the machine learning algorithms.
            eval_method: The name of the method to evaluate the quality measure.
            samples: The indices of the learning samples to consider.
                Other indices are neither used for training nor for testing.
                If None, use all the samples.
            **measure_options: The options for the method
                to evaluate the quality measure.
                The option 'multioutput' will be set to False.

        Raises:
            ValueError: If the unsupported "multioutput" option is enabled.
        """
        self.dataset = dataset
        self.measure = measure
        self.measure_options = dict(
            samples=samples, method=eval_method, **measure_options
        )
        self.factory = MLAlgoFactory()

        self.candidates = []

        if "multioutput" in measure_options and measure_options["multioutput"]:
            raise ValueError(
                "MLAlgoSelection does not support multioutput. "
                "The measure shall return one value."
            )
        self.measure_options["multioutput"] = False

    def add_candidate(
        self,
        name,  # type:str
        calib_space=None,  # type: Optional[DesignSpace]
        calib_algo=None,  # type: Optional[ScenarioInputDataType]
        **option_lists
    ):  # type: (...) -> None
        """Add a machine learning algorithm candidate.

        Args:
            name: The name of a machine learning algorithm.
            calib_space: The design space
                defining the parameters to be calibrated
                with a :class:`.MLAlgoCalibration`.
                If None, do not perform calibration.
            calib_algo: The name and the parameters
                of the optimization algorithm,
                e.g. {"algo": "fullfact", "n_samples": 10}).
                If None, do not perform calibration.
            **option_lists: The parameters
                for the machine learning algorithm candidate.
                Each parameter has to be enclosed within a list.
                The list may contain different values
                to try out for the given parameter,
                or only one.

        Examples:
            >>> selector.add_candidate(
            >>>     "LinearRegression",
            >>>     penalty_level=[0, 0.1, 1, 10, 20],
            >>>     l2_penalty_ratio=[0, 0.5, 1],
            >>>     fit_intercept=[True],
            >>> )
        """
        keys, values = option_lists.keys(), option_lists.values()

        # Set initial quality to the worst possible value
        if self.measure.SMALLER_IS_BETTER:
            quality = float("inf")
        else:
            quality = -float("inf")

        for prodvalues in product(*values):
            params = dict(zip(keys, prodvalues))
            if not calib_space:
                algo_new = self.factory.create(name, data=self.dataset, **params)
                quality_new = self.measure(algo_new).evaluate(**self.measure_options)
            else:
                calib = MLAlgoCalibration(
                    name,
                    self.dataset,
                    calib_space.variables_names,
                    calib_space,
                    self.measure,
                    self.measure_options,
                    **params
                )
                calib.execute(calib_algo)
                algo_new = calib.optimal_algorithm
                quality_new = calib.optimal_criterion

            if self.measure.is_better(quality_new, quality):
                algo = algo_new
                quality = quality_new

        # Learn on the entire dataset if a cross-validation scheme was used
        algo.learn(self.measure_options["samples"])

        self.candidates.append((algo, quality))

    def select(
        self,
        return_quality=False,  # type:bool
    ):  # type: (...) -> Union[MLAlgo,Tuple[MLAlgo,float]]
        """Select the best model.

        The model is chosen through a grid search
        over candidates and their options,
        as well as an eventual optimization over the parameters
        in the calibration space.

        Args:
            return_quality: If True, the quality of the best model will be returned.

        Returns:
            The best model and its quality if required.
        """
        candidate = self.candidates[0]
        for new_candidate in self.candidates[1:]:
            if self.measure.is_better(new_candidate[1], candidate[1]):
                candidate = new_candidate
        if not return_quality:
            candidate = candidate[0]
        return candidate



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scenario which drivers are Design of Experiments
************************************************
"""
from __future__ import division, unicode_literals

import logging

from gemseo.algos.doe.doe_factory import DOEFactory
from gemseo.core.scenario import Scenario

# The detection of formulations requires to import them,
# before calling get_formulation_from_name
LOGGER = logging.getLogger(__name__)


class DOEScenario(Scenario):
    """Design of Experiments scenario, based on MDO scenario but with a DOE driver.

    The main differences between Scenario and MDOScenario are the allowed
    inputs
    in the MDOScenario.json, which differs from DOEScenario.json, at least
    on the driver names

    MDO Problem description: links the disciplines and the formulation
    to create an optimization problem.
    Use the class by instantiation.
    Create your disciplines beforehand.

    Specify the formulation by giving the class name such as the string
    "MDF"

    The reference_input_data is the typical input data dict that is provided
    to the run method of the disciplines

    Specify the objective function name, which must be an output
    of a discipline of the scenario, with the "objective_name" attribute

    If you want to add additional design constraints,
    use the add_user_defined_constraint method

    To view the results, use the "post_process" method after execution.
    You can view:

    - the design variables history, the objective value, the constraints,
      by using: scenario.post_process("OptHistoryView", show=False, save=True)
    - Quadratic approximations of the functions close to the
      optimum, when using gradient based algorithms, by using:
      scenario.post_process("QuadApprox", method="SR1", show=False,
      save=True, function="my_objective_name",
      file_path="appl_dir")
    - Self Organizing Maps of the design space, by using:
      scenario.post_process("SOM", save=True, file_path="appl_dir")

    To list post processings on your setup,
    use the method :attr:`.Scenario.posts`.
    For more details on their options, go to the **gemseo.post** package.
    """

    # Constants for input variables in json schema
    N_SAMPLES = "n_samples"
    EVAL_JAC = "eval_jac"
    SEED = "seed"

    def __init__(
        self,
        disciplines,
        formulation,
        objective_name,
        design_space,
        name=None,
        **formulation_options
    ):
        """Constructor, initializes the DOE scenario Objects instantiation and checks
        are made before run intentionally.

        :param disciplines: the disciplines of the scenario
        :param formulation: the formulation name,
            the class name of the formulation in gemseo.formulations
        :param objective_name: the objective function name
        :param design_space: the design space
        :param name: scenario name
        :param formulation_options: options for creation of the formulation
        """
        # This loads the right json grammars from class name
        super(DOEScenario, self).__init__(
            disciplines,
            formulation,
            objective_name,
            design_space,
            name,
            **formulation_options
        )
        self.seed = 0
        self.default_inputs = {self.EVAL_JAC: False, self.ALGO: "lhs"}

    def _init_algo_factory(self):
        """Initalizes the algorithms factory."""
        self._algo_factory = DOEFactory()

    def _run_algorithm(self):
        """Runs the DOE algo."""
        self.seed += 1
        problem = self.formulation.opt_problem
        algo_name = self.local_data[self.ALGO]
        n_samples = self.local_data.get(self.N_SAMPLES)
        options = self.local_data.get(self.ALGO_OPTIONS)
        if options is None:
            options = {}
        if self.N_SAMPLES in options:
            LOGGER.warning(
                "Double definition of algorithm option n_samples, " "keeping value: %s",
                n_samples,
            )
            options.pop(self.N_SAMPLES)

        if self.ALGO_OPTIONS in self.local_data:
            options = self.local_data[self.ALGO_OPTIONS]
        lib = self._algo_factory.create(algo_name)
        lib.init_options_grammar(algo_name)
        if self.SEED in lib.opt_grammar.get_data_names():
            if self.SEED not in options:
                options[self.SEED] = self.seed
        self.optimization_result = lib.execute(problem, n_samples=n_samples, **options)
        return self.optimization_result

    def _run(self):
        """Execute the scenario and run the optimization problems."""
        LOGGER.info(" ")
        LOGGER.info("*** Start DOE Scenario execution ***")
        LOGGER.info("%s", repr(self))
        self._run_algorithm()
        LOGGER.info("*** DOE Scenario run terminated ***")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The design space for the MDO problem proposed by Sellar et al. in.

Sellar, R., Batill, S., & Renaud, J. (1996). Response surface based, concurrent subspace
optimization for multidisciplinary system design. In 34th aerospace sciences meeting and
exhibit (p. 714).
"""
from __future__ import division, unicode_literals

from typing import Tuple

from numpy import array, ndarray

from gemseo.algos.design_space import DesignSpace
from gemseo.problems.sellar.sellar import X_LOCAL, X_SHARED, Y_1, Y_2


class SellarDesignSpace(DesignSpace):
    """The design space for the MDO problem proposed by Sellar et al (1996).

    It is composed of:
    - :math:`x_{local}` belonging to :math:`[0., 10.]`,
    - :math:`x_{shared,1}` belonging to :math:`[-10., 10.]`,
    - :math:`x_{shared,2}` belonging to :math:`[0., 10.]`,
    - :math:`y_1` belonging to :math:`[-100., 100.]`,
    - :math:`y_2` belonging to :math:`[-100., 100.]`.

    This design space is initialized with the initial solution:

    - :math:`x_{local}=1`,
    - :math:`x_{shared,1}=4`,
    - :math:`x_{shared,2}=3`,
    - :math:`y_1=1`,
    - :math:`y_2=1`.
    """

    def __init__(
        self,
        dtype="complex128",  # type: str
    ):  # type: (...) -> None
        """
        Args:
            dtype: The type of the variables defined in the design space.
        """
        super(SellarDesignSpace, self).__init__()

        x_local, x_shared, y_1, y_2 = self.__get_initial_solution(dtype)
        self.add_variable(X_LOCAL, 1, l_b=0.0, u_b=10.0, value=x_local)
        self.add_variable(X_SHARED, 2, l_b=(-10, 0.0), u_b=(10.0, 10.0), value=x_shared)
        self.add_variable(Y_1, 1, l_b=-100.0, u_b=100.0, value=y_1)
        self.add_variable(Y_2, 1, l_b=-100.0, u_b=100.0, value=y_2)

    @staticmethod
    def __get_initial_solution(
        dtype,  # type: str
    ):  # type: (...) -> Tuple[ndarray]
        """Return an initial solution for the MDO problem.

        Args:
            dtype: The type of the variables defined in the design space.

        Returns:
            An initial solution for both local design variables,
            shared design variables and coupling variables.
        """
        x_local = array([1.0], dtype=dtype)
        x_shared = array([4.0, 3.0], dtype=dtype)
        y_1 = array([1.0], dtype=dtype)
        y_2 = array([1.0], dtype=dtype)
        return x_local, x_shared, y_1, y_2



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Computation of tolerance intervals from a data-fitted Weibull distribution."""

import openturns as ot
from numpy import exp, log

from gemseo.uncertainty.statistics.tolerance_interval.distribution import (
    ToleranceInterval,
)


class WeibullToleranceInterval(ToleranceInterval):
    """Computation of tolerance intervals from a data-fitted Weibull distribution.

    The formulae come from the R library *tolerance* [1]_.

    .. [1] Derek S. Young,
           *tolerance: An R Package for Estimating Tolerance Intervals*,
           Journal of Statistical Software, 36(5), 2010
    """

    def __init__(
        self,
        size,  # type: int
        scale,  # type: float
        shape,  # type: float
        location,  # type: float
    ):  # type:(...) -> None
        # noqa: D205 D212 D415
        """
        Args:
            scale: The estimation of the scale of the Weibull distribution.
            shape: The estimation of the shape of the Weibull distribution.
            location: The estimation of the location of the Weibull distribution.
        """
        super(WeibullToleranceInterval, self).__init__(size)
        self.__scale = scale
        self.__shape = shape
        self.__location = location

    @staticmethod
    def __lambda_function(
        value,  # type: float
    ):  # type: (...) -> float
        """Compute the natural logarithm of the opposite natural logarithm.

        Args:
            value: The value to be transformed.

        Returns:
            The transformed value.
        """
        return log(-log(value))

    def _compute_lower_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        xi_ = log(self.__scale)
        delta = 1.0 / self.__shape
        offset = -(size ** 0.5) * self.__lambda_function(coverage)
        student = ot.Student(size - 1, offset)
        bound = xi_ - delta * student.computeQuantile(1 - alpha)[0] / (size - 1) ** 0.5
        return exp(bound) + self.__location

    def _compute_upper_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        xi_ = log(self.__scale)
        delta = 1.0 / self.__shape
        offset = -(size ** 0.5) * self.__lambda_function(1 - coverage)
        student = ot.Student(size - 1, offset)
        bound = xi_ - delta * student.computeQuantile(alpha)[0] / (size - 1) ** 0.5
        return exp(bound) + self.__location


class WeibullMinToleranceInterval(WeibullToleranceInterval):
    """Computation of tolerance intervals from a data-fitted Weibull distribution.

    The formulae come from the R library *tolerance* [1]_.

    .. [1] Derek S. Young,
           *tolerance: An R Package for Estimating Tolerance Intervals*,
           Journal of Statistical Software, 36(5), 2010
    """



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Services for handling Matplotlib figures, e.g. save and show."""

from typing import Union

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from gemseo.utils.py23_compat import Path


def save_show_figure(
    fig,  # type: Figure
    show,  # type:bool
    file_path,  # type: Union[str,Path]
):  # type: (...) -> None
    """Save or show a Matplotlib figure.

    Args:
        fig: The Matplotlib figure to be saved or shown.
        show: If True, display the Matplotlib figure.
        file_path: The file path to save the Matplotlib figure.
            If None, do not save the figure.
    """
    save = file_path is not None

    if save:
        fig.savefig(str(file_path), bbox_inches="tight")

    if show:
        try:
            plt.show(fig)
        except TypeError:
            plt.show()

    if save or show:
        plt.close(fig)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Here is the baseclass to measure the quality of machine learning algorithms.

The concept of quality measure is implemented with the :class:`.MLQualityMeasure` class.
"""
from __future__ import division, unicode_literals

from typing import List, NoReturn, Optional, Union

import six
from custom_inherit import DocInheritMeta
from numpy import arange, ndarray

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import MLAlgo

OptionType = Optional[Union[List[int], bool, int, Dataset]]


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
    )
)
class MLQualityMeasure(object):
    """An abstract quality measure for machine learning algorithms.

    Attributes:
        algo (MLAlgo): The machine learning algorithm.
    """

    LEARN = "learn"
    TEST = "test"
    LOO = "loo"
    KFOLDS = "kfolds"
    BOOTSTRAP = "bootstrap"

    SMALLER_IS_BETTER = True  # To be overwritten in inheriting classes

    def __init__(
        self,
        algo,  # type: MLAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm.
        """
        self.algo = algo

    def evaluate(
        self,
        method=LEARN,  # type: str
        samples=None,  # type: Optional[List[int]]
        **options  # type: Optional[OptionType]
    ):  # type: (...) -> Union[float,ndarray]
        """Evaluate the quality measure.

        Args:
            method: The name of the method
                to evaluate the quality measure.
            samples: The indices of the learning samples.
                If None, use the whole learning dataset.
            **options: The options of the estimation method (e.g. 'test_data' for
            the 'test' method, 'n_replicates' for the bootstrap one, ...)

        Returns:
            The value of the quality measure.

        Raises:
            ValueError: If the name of the method is unknown.
        """
        if method == self.LEARN:
            evaluation = self.evaluate_learn(samples=samples, **options)
        elif method == self.TEST:
            evaluation = self.evaluate_test(samples=samples, **options)
        elif method == self.LOO:
            evaluation = self.evaluate_loo(samples=samples, **options)
        elif method == self.KFOLDS:
            evaluation = self.evaluate_kfolds(samples=samples, **options)
        elif method == self.BOOTSTRAP:
            evaluation = self.evaluate_bootstrap(samples=samples, **options)
        else:
            raise ValueError("The method '{}' is not available.".format(method))
        return evaluation

    def evaluate_learn(
        self,
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> NoReturn
        """Evaluate the quality measure using the learning dataset.

        Args:
            samples: The indices of the learning samples.
                If None, use the whole learning dataset.
            multioutput: If True, return the quality measure for each
                output component. Otherwise, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError

    def evaluate_test(
        self,
        test_data,  # type:Dataset
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> NoReturn
        """Evaluate the quality measure using a test dataset.

        Args:
            dataset: The test dataset.
            samples: The indices of the learning samples.
                If None, use the whole learning dataset.
            multioutput: If True, return the quality measure for each
                output component. Otherwise, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError

    def evaluate_loo(
        self,
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        """Evaluate the quality measure using the leave-one-out technique.

        Args:
            samples: The indices of the learning samples.
                If None, use the whole learning dataset.
            multioutput: If True, return the quality measure for each
                output component. Otherwise, average these measures.

        Returns:
            The value of the quality measure.
        """
        n_samples = self.algo.learning_set.n_samples
        return self.evaluate_kfolds(n_folds=n_samples, multioutput=multioutput)

    def evaluate_kfolds(
        self,
        n_folds=5,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> NoReturn
        """Evaluate the quality measure using the k-folds technique.

        Args:
            n_folds: The number of folds.
            samples: The indices of the learning samples.
                If None, use the whole learning dataset.
            multioutput: If True, return the quality measure for each
                output component. Otherwise, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError

    def evaluate_bootstrap(
        self,
        n_replicates=100,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> NoReturn
        """Evaluate the quality measure using the bootstrap technique.

        Args:
            n_replicates: The number of bootstrap replicates.
            samples: The indices of the learning samples.
                If None, use the whole learning dataset.
            multioutput: If True, return the quality measure for each
                output component. Otherwise, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError

    @classmethod
    def is_better(
        cls,
        val1,  # type: float
        val2,  # type: float
    ):  # type: (...) -> bool
        """Compare the quality between two values.

        This methods returns True if the first one is better than the second one.

        For most measures, a smaller value is "better" than a larger one (MSE
        etc.). But for some, like an R2-measure, higher values are better than
        smaller ones. This comparison method correctly handles this,
        regardless of the type of measure.

        Args:
            val1: The value of the first quality measure.
            val2: The value of the second quality measure.

        Returns:
            Whether val1 is of better quality than val2.
        """
        if cls.SMALLER_IS_BETTER:
            result = val1 < val2
        else:
            result = val1 > val2
        return result

    def _assure_samples(
        self,
        samples,  # type: List[int]
    ):  # type: (...) -> Union[List[int],ndarray]
        """Get the list of all samples if samples is None.

        Args:
            samples: The list of samples. Can also be None.

        Returns:
            The samples.
        """
        if samples is None:
            samples = arange(self.algo.learning_set.n_samples)
        return samples



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Surrogate discipline baseclass
******************************
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.discipline import MDODiscipline
from gemseo.core.jacobian_assembly import JacobianAssembly
from gemseo.mlearning.regression.factory import RegressionModelFactory
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.string_tools import MultiLineString, pretty_repr

LOGGER = logging.getLogger(__name__)


class SurrogateDiscipline(MDODiscipline):
    """Surrogate discipline class."""

    def __init__(
        self,
        surrogate,
        data=None,
        transformer=MLRegressionAlgo.DEFAULT_TRANSFORMER,
        disc_name=None,
        default_inputs=None,
        input_names=None,
        output_names=None,
        **parameters
    ):
        """Constructor.

        :param surrogate: name of the surrogate model algorithm.
        :type surrogate: str or MLRegressionAlgo
        :param Dataset data: dataset to train the surrogate. If None,
            assumes that the surrogate is trained. Default: None.
        :param dict(Transformer) transformer: transformation strategy for data groups.
            If None, do not scale data.
            Default: MLRegressionAlgo.DEFAULT_TRANSFORMER,
            which is a min/max scaler applied to the inputs
            and a min/max scaler applied to the outputs.
        :param str disc_name: name of the surrogate discipline.
            If None, use surrogate.ABBR + data.name . Default: None
        :param dict default_inputs: default inputs. If None, use the center of the
            learning input space. Default: None.
        :param list(str) input_names: list of input names.
            If None, use all inputs. Default: None.
        :param list(str) output_names: list of output names.
            If None, use all outputs. Default: None.
        :param parameters: surrogate model parameters.
        """
        if isinstance(surrogate, MLRegressionAlgo):
            self.regression_model = surrogate
            name = self.regression_model.learning_set.name
        elif data is None:
            raise ValueError("data is required to train the surrogate model.")
        else:
            factory = RegressionModelFactory()
            self.regression_model = factory.create(
                surrogate,
                data=data,
                transformer=transformer,
                input_names=input_names,
                output_names=output_names,
                **parameters
            )
            name = self.regression_model.ABBR + "_" + data.name
        disc_name = disc_name or name
        if not self.regression_model.is_trained:
            self.regression_model.learn()
            msg = MultiLineString()
            msg.add("Build the surrogate discipline: {}", disc_name)
            msg.indent()
            msg.add("Dataset name: {}", data.name)
            msg.add("Dataset size: {}", data.length)
            msg.add("Surrogate model: {}", self.regression_model.__class__.__name__)
            LOGGER.info("%s", msg)
        if not name.startswith(self.regression_model.ABBR):
            disc_name = self.regression_model.ABBR + "_" + disc_name
        msg = MultiLineString()
        msg.add("Use the surrogate discipline: {}", disc_name)
        msg.indent()
        super(SurrogateDiscipline, self).__init__(disc_name)
        self._initialize_grammars(input_names, output_names)
        msg.add("Inputs: {}", pretty_repr(self.get_input_data_names()))
        msg.add("Outputs: {}", pretty_repr(self.get_output_data_names()))
        self._set_default_inputs(default_inputs)
        self.add_differentiated_inputs()
        self.add_differentiated_outputs()
        try:
            self.regression_model.predict_jacobian(self.default_inputs)
            self.linearization_mode = JacobianAssembly.AUTO_MODE
            msg.add("Jacobian: use surrogate model jacobian")
        except NotImplementedError:
            self.linearization_mode = self.FINITE_DIFFERENCES
            msg.add("Jacobian: use finite differences")
        LOGGER.info("%s", msg)

    def __repr__(self):
        model = self.regression_model.__class__.__name__
        data_name = self.regression_model.learning_set.name
        length = len(self.regression_model.learning_set)
        inputs = sorted(self.regression_model.input_names)
        outputs = sorted(self.regression_model.output_names)
        arguments = [
            "name={}".format(self.name),
            "algo={}".format(model),
            "data={}".format(data_name),
            "size={}".format(length),
            "inputs=[{}]".format(pretty_repr(inputs)),
            "outputs=[{}]".format(pretty_repr(outputs)),
            "jacobian={}".format(self.linearization_mode),
        ]
        msg = "SurrogateDiscipline({})".format(", ".join(arguments))
        return msg

    def __str__(self):
        data_name = self.regression_model.learning_set.name
        length = len(self.regression_model.learning_set)
        msg = MultiLineString()
        msg.add("Surrogate discipline: {}", self.name)
        msg.indent()
        msg.add("Dataset name: {}", data_name)
        msg.add("Dataset size: {}", length)
        msg.add("Surrogate model: {}", self.regression_model.__class__.__name__)
        inputs = sorted(self.regression_model.input_names)
        outputs = sorted(self.regression_model.output_names)
        msg.add("Inputs: {}", pretty_repr(inputs))
        msg.add("Outputs: {}", pretty_repr(outputs))
        return str(msg)

    def _initialize_grammars(self, input_names=None, output_names=None):
        """Initializes the inputs and outputs grammars from data."""
        learning_set = self.regression_model.learning_set
        in_grp = learning_set.INPUT_GROUP
        out_grp = learning_set.OUTPUT_GROUP
        if input_names is None:
            inputs = learning_set.get_data_by_group(in_grp)[0, :]
            input_names = learning_set.get_names(in_grp)
        else:
            inputs = learning_set.get_data_by_names(input_names, False)[0, :]
        if output_names is None:
            outputs = learning_set.get_data_by_group(out_grp)[0, :]
            output_names = learning_set.get_names(out_grp)
        else:
            outputs = learning_set.get_data_by_names(output_names, False)[0, :]
        inputs = DataConversion.array_to_dict(inputs, input_names, learning_set.sizes)
        outputs = DataConversion.array_to_dict(
            outputs, output_names, learning_set.sizes
        )
        self.input_grammar.initialize_from_base_dict(inputs)
        self.output_grammar.initialize_from_base_dict(outputs)

    def _set_default_inputs(self, default_inputs=None):
        """Set default inputs either from the center of the learning input space or from
        user specification.

        :param dict default_inputs: user default inputs.
            If None, use the learning input space center. Default: None.
        """
        if default_inputs is None:
            self._default_inputs = self.regression_model.input_space_center
        else:
            self._default_inputs = default_inputs

    def _run(self):
        input_data = self.get_input_data()
        output_data = self.regression_model.predict(input_data)
        output_data = {key: val.flatten() for key, val in output_data.items()}
        self.local_data.update(output_data)

    def _compute_jacobian(self, inputs=None, outputs=None):
        input_data = self.get_input_data()
        self._init_jacobian(inputs, outputs)
        self.jac = self.regression_model.predict_jacobian(input_data)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Benoit Pauwels
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Wrapper for the Generic Tool for Optimization (GTOpt) of pSeven Core."""

from __future__ import unicode_literals

from enum import Enum
from math import sqrt
from typing import Any, Dict, List, Mapping, Optional, Union

from da import p7core
from numpy import ndarray

from gemseo.algos.opt.core.pseven_problem_adapter import CostType, PSevenProblem
from gemseo.algos.opt.opt_lib import OptimizationLibrary
from gemseo.algos.opt_result import OptimizationResult
from gemseo.core.function import MDOLinearFunction, MDOQuadraticFunction
from gemseo.utils.base_enum import CamelCaseEnum


class DiffScheme(CamelCaseEnum):
    """The differentiation schemes of pSeven."""

    FIRST_ORDER = 0
    SECOND_ORDER = 1
    # GTOpt switches between first and second order
    # depending on the estimated distance to optimality.
    ADAPTIVE = 2
    AUTO = 3  # GTOpt is left free to choose.


class DiffType(CamelCaseEnum):
    """The differentiation types of pSeven."""

    NUMERICAL = 0  # conventional numerical differentiation
    # framed simplex-based derivatives, recommended for noisy problems
    FRAMED = 1
    AUTO = 2  # GTOpt is left free to choose.


class GlobalMethod(Enum):
    """The globalization methods of pSeven.

    A item is represented with the name of its key.
    """

    RL = 0  # random linkages
    PM = 1  # plain multistart
    MS = 2  # surrogate model-based multistart

    def __str__(self):  # type: (...) -> str
        return self.name


class LogLevel(CamelCaseEnum):
    """The logging levels of pSeven."""

    DEBUG = 0
    INFO = 1
    WARN = 2
    ERROR = 3
    FATAL = 4


class Smoothness(CamelCaseEnum):
    """The functions smoothness levels of pSeven."""

    SMOOTH = 0  # all functions are to be considered smooth
    NOISY = 1  # at least one function is noisy with noise level at most 10 %
    AUTO = 2  # GTOpt is free to assume whatever seems appropriate.


class PSevenOpt(OptimizationLibrary):
    """Interface for the Generic Tool for Optimization (GTOpt) of pSeven Core."""

    LIB_COMPUTE_GRAD = True
    OPTIONS_MAP = {
        "constraints_smoothness": "GTOpt/ConstraintsSmoothness",
        "global_phase_intensity": "GTOpt/GlobalPhaseIntensity",
        "max_expensive_func_iter": "GTOpt/MaximumExpensiveIterations",
        "max_func_iter": "GTOpt/MaximumIterations",
        "objectives_smoothness": "GTOpt/ObjectivesSmoothness",
        "deterministic": "GTOpt/Deterministic",
        "log_level": "GTOpt/LogLevel",
        "verbose_log": "GTOpt/VerboseOutput",
        "max_threads": "GTOpt/MaxParallel",
        "seed": "GTOpt/Seed",
        "time_limit": "GTOpt/TimeLimit",
        "grad_tol": "GTOpt/GradientTolerance",
        "grad_tol_is_abs": "GTOpt/AbsoluteGradientTolerance",
        "max_batch_size": "GTOpt/BatchSize",
        "detect_nan_clusters": "GTOpt/DetectNaNClusters",
        "diff_scheme": "GTOpt/DiffScheme",
        "diff_type": "GTOpt/DiffType",
        "diff_step": "GTOpt/NumDiffStepSize",
        "ensure_feasibility": "GTOpt/EnsureFeasibility",
        "local_search": "GTOpt/LocalSearch",
        "restore_analytic_func": "GTOpt/RestoreAnalyticResponses",
    }

    # Governing methods
    # Not yet used.
    # SO = "SO"
    # MO = "MO"
    __SBO = "SBO"

    # Local methods
    __FD = "FD"
    __MOM = "MoM"
    __NCG = "NCG"
    __NLS = "NLS"
    __POWELL = "Powell"
    __QP = "QP"
    __SQP = "SQP"
    __SQ2P = "S2P"
    __LOCAL_METHODS = (__FD, __MOM, __NCG, __NLS, __POWELL, __QP, __SQP, __SQ2P)

    __WEBSITE = "https://datadvance.net/product/pseven/manual/"

    def __init__(self):  # type: (...) -> None # noqa: D107
        super(PSevenOpt, self).__init__()
        self.lib_dict = {
            "PSEVEN": {
                self.INTERNAL_NAME: "pSeven",
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "pSeven's Generic Tool for Optimization (GTOpt).",
                self.WEBSITE: self.__WEBSITE,
            },
            "PSEVEN_FD": {
                self.INTERNAL_NAME: self.__FD,
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "pSeven's feasible direction method.",
                self.WEBSITE: self.__WEBSITE,
            },
            "PSEVEN_MOM": {
                self.INTERNAL_NAME: self.__MOM,
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "pSeven's method of multipliers.",
                self.WEBSITE: self.__WEBSITE,
            },
            "PSEVEN_NCG": {
                self.INTERNAL_NAME: self.__NCG,
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "pSeven's nonlinear conjugate gradient method.",
                self.WEBSITE: self.__WEBSITE,
            },
            "PSEVEN_NLS": {
                self.INTERNAL_NAME: self.__NLS,
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "pSeven's nonlinear simplex method.",
                self.WEBSITE: self.__WEBSITE,
            },
            "PSEVEN_POWELL": {
                self.INTERNAL_NAME: self.__POWELL,
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "pSeven's Powell conjugate direction method.",
                self.WEBSITE: self.__WEBSITE,
            },
            "PSEVEN_QP": {
                self.INTERNAL_NAME: self.__QP,
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "pSeven's quadratic programming method.",
                self.WEBSITE: self.__WEBSITE,
            },
            "PSEVEN_SQP": {
                self.INTERNAL_NAME: self.__SQP,
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "pSeven's sequential quadratic programming method.",
                self.WEBSITE: self.__WEBSITE,
            },
            "PSEVEN_SQ2P": {
                self.INTERNAL_NAME: self.__SQ2P,
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: True,
                self.HANDLE_INEQ_CONS: True,
                self.DESCRIPTION: "pSeven's sequential quadratic constrained quadratic "
                "programming method.",
                self.WEBSITE: self.__WEBSITE,
            },
        }

    def _get_options(
        self,
        max_iter=99,  # type: int
        evaluation_cost_type=None,  # type: Optional[Mapping[str, CostType]]
        expensive_evaluations=None,  # type: Optional[Mapping[str, int]]
        sample_x=None,  # type: Optional[Union[List[float],List[ndarray]]]
        sample_f=None,  # type: Optional[Union[List[float],List[ndarray]]]
        sample_c=None,  # type: Optional[Union[List[float],List[ndarray]]]
        constraints_smoothness=Smoothness.AUTO,  # type: Smoothness
        global_phase_intensity=None,  # type: Optional[float]
        max_expensive_func_iter=0,  # type: int
        max_func_iter=0,  # type: int
        objectives_smoothness=Smoothness.AUTO,  # type: Smoothness
        deterministic=None,  # type: Optional[bool]
        log_level=LogLevel.INFO,  # type: LogLevel
        verbose_log=False,  # type: bool
        max_threads=0,  # type: int
        seed=100,  # type: int
        time_limit=0,  # type: int
        grad_tol=1e-5,  # type: float
        grad_tol_is_abs=False,  # type: bool
        max_batch_size=0,  # type: int
        detect_nan_clusters=True,  # type: bool
        diff_scheme=DiffScheme.AUTO,  # type: DiffScheme
        diff_type=DiffType.AUTO,  # type: DiffType
        diff_step=1.1920929e-06,  # type: float
        ensure_feasibility=False,  # type: bool
        local_search=False,  # type: bool
        restore_analytic_func=True,  # type: bool
        globalization_method=None,  # type: Optional[GlobalMethod]
        surrogate_based=None,  # type: Optional[bool]
        **kwargs  # type: Any
    ):  # type: (...) -> Dict
        """Set the default options values.

        Args:
            max_iter: The maximum number of evaluations.
            evaluation_cost_type: The evaluation cost type of each function of the
                problem.
                If None, the evaluation cost types default to "Cheap".
            expensive_evaluations: The maximal number of expensive evaluations for
                each function of the problem. By default, set automatically by pSeven.
            sample_x: A sample of design points (in addition to the problem initial
                design).
            sample_f: The objectives values at the design points of the sample.
            sample_c: The constraints values at the design points of the sample.
            constraints_smoothness: The assumed smoothness of the constraints functions.
            global_phase_intensity: The configuration of global searching algorithms.
                This option has different meanings for expensive and non-expensive
                optimization problems. Refer to the pSeven Core API documentation.
                Defaults to "Auto".
            max_expensive_func_iter: The maximum number of evaluations for each
                expensive response, excluding the evaluations of initial guesses.
            max_func_iter: The maximum number of evaluations for any response,
                including the evaluations of initial guesses.
            objectives_smoothness: The assumed smoothness of the objective functions.
            deterministic: Whether to require optimization process to be reproducible
                using the passed seed value. Defaults to "Auto".
            log_level: The minimum log level.
            verbose_log: Whether to enable verbose logging.
            max_threads: The maximum number of parallel threads to use when solving.
            seed: The random seed for deterministic mode.
            time_limit: The maximum allowed time to solve a problem in seconds.
                Defaults to 0, unlimited.
            grad_tol: The tolerance on the infinity-norm of the gradient (or optimal
                descent for constrained and multi-objective problems) at which
                optimization stops.
                If 'gradient_tol_is_abs' is False then the tolerance is relative to
                the infinity-norm of the current objectives values; otherwise the
                tolerance is absolute.
                The value 0.0 deactivate the gradient-based stopping criterion.
            grad_tol_is_abs: Whether 'grad_tol' should be regarded as an absolute
                tolerance. See 'grad_tol' for details.
            max_batch_size: The maximum number of points in an evaluation batch.
                The (default) value 0 allows the optimizer to use any batch size.
            detect_nan_clusters: Whether to detect and avoid design space areas that
                yield NaN values (for at least one function).
                This option has no effect in the absence of "expensive" functions.
            diff_scheme: The order of the differentiation scheme (when the analytic
                derivatives are unavailable).
            diff_type: The strategy for differentiation (when the analytic
                derivatives are unavailable).
            diff_step: The numerical differentiation step size.
            ensure_feasibility: Whether to restrict the evaluations of the objectives
                to feasible designs only.
            local_search: Whether to force the surrogate models to explore the design
                space locally near the current optimum,
                or to disable the local search and let the surrogate models explore the
                whole design space.
            restore_analytic_func: Whether to restore the analytic forms of the
                linear and quadratic functions.
                Once the analytic forms are restored the original functions will not
                be evaluated anymore.
            globalization_method: The globalization method.
                If None, set automatically depending on the problem.
            surrogate_based: Whether to use surrogate models.
                If None, set automatically depending on the problem.
            **kwargs: Other driver options.

        Returns:
            The processed options.

        Raises:
            ValueError: If the value for one the following option is invalid:
                objectives smoothness,
                constraints smoothness,
                logging level,
                differentiation scheme,
                differentiation type,
                globalization method.
        """
        # Check the options
        self.__check_pseven_options(
            evaluation_cost_type,
            expensive_evaluations,
            constraints_smoothness,
            objectives_smoothness,
            log_level,
            diff_scheme,
            diff_type,
            globalization_method,
        )

        # Process the options
        if globalization_method is not None:
            globalization_method = str(globalization_method)

        processed_options = self._process_options(
            max_iter=max_iter,
            evaluation_cost_type=evaluation_cost_type,
            expensive_evaluations=expensive_evaluations,
            sample_x=sample_x,
            sample_f=sample_f,
            sample_c=sample_c,
            constraints_smoothness=str(constraints_smoothness),
            global_phase_intensity=global_phase_intensity,
            max_expensive_func_iter=max_expensive_func_iter,
            max_func_iter=max_func_iter,
            objectives_smoothness=str(objectives_smoothness),
            deterministic=deterministic,
            log_level=str(log_level),
            verbose_log=verbose_log,
            max_threads=max_threads,
            seed=seed,
            time_limit=time_limit,
            grad_tol=grad_tol,
            grad_tol_is_abs=grad_tol_is_abs,
            max_batch_size=max_batch_size,
            detect_nan_clusters=detect_nan_clusters,
            diff_scheme=str(diff_scheme),
            diff_type=str(diff_type),
            diff_step=diff_step,
            ensure_feasibility=ensure_feasibility,
            local_search=local_search,
            restore_analytic_func=restore_analytic_func,
            globalization_method=globalization_method,
            surrogate_based=surrogate_based,
            **kwargs
        )

        gtopt_local_search = "GTOpt/LocalSearch"

        if processed_options[gtopt_local_search]:
            processed_options[gtopt_local_search] = "Forced"
        else:
            processed_options[gtopt_local_search] = "Disabled"

        # Disable pSeven's internal stopping criterion based on successive designs
        processed_options["GTOpt/CoordinateTolerance"] = 0.0

        # Disable pSeven's internal stopping criterion based on successive objectives
        processed_options["GTOpt/ObjectiveTolerance"] = 0.0

        # Set the tolerance on the constraints (N.B. relative to the infinity-norm)
        processed_options[
            "GTOpt/ConstraintsTolerance"
        ] = self.__compute_constraints_tolerance()

        # Set the pSeven's techniques
        self.__set_pseven_techniques(processed_options)

        return processed_options

    def __check_pseven_options(
        self,
        evaluation_cost_type,  # type: Mapping[str, CostType]
        expensive_evaluations,  # type: Mapping[str, int]
        constraints_smoothness,  # type: Smoothness
        objectives_smoothness,  # type: Smoothness
        log_level,  # type: LogLevel
        diff_scheme,  # type: DiffScheme
        diff_type,  # type: DiffType
        globalization_method,  # type: GlobalMethod
    ):  # type: (...) -> None
        """Check pSeven's options.

        Args:
            evaluation_cost_type: The evaluation cost type of each function of the
                problem.
                If None, the evaluation cost types default to "Cheap".
            expensive_evaluations: The maximal number of expensive evaluations for
                each function of the problem.
                 If None, set automatically by pSeven.
            constraints_smoothness: The assumed smoothness of the constraints functions.
            objectives_smoothness: The assumed smoothness of the objective functions.
            log_level: The minimum log level.
            diff_scheme: The order of the differentiation scheme (when the analytic
                derivatives are unavailable).
            diff_type: The strategy for differentiation (when the analytic derivatives
                are unavailable).
            globalization_method: The globalization method.

        Raises:
            ValueError: If the value for one the following option is invalid:
                objectives smoothness,
                constraints smoothness,
                logging level,
                differentiation scheme,
                differentiation type,
                globalization method.
        """
        if evaluation_cost_type is not None:
            self.__check_evaluation_cost_type(evaluation_cost_type)

        if expensive_evaluations is not None:
            self.__check_expensive_evaluations(expensive_evaluations)

        if not isinstance(objectives_smoothness, Smoothness):
            raise ValueError(
                "Unknown objectives smoothness: {}".format(objectives_smoothness)
            )

        if not isinstance(constraints_smoothness, Smoothness):
            raise ValueError(
                "Unknown constraints smoothness: {}".format(constraints_smoothness)
            )

        if not isinstance(log_level, LogLevel):
            raise ValueError("Unknown log level: {}".format(log_level))

        if not isinstance(diff_scheme, DiffScheme):
            raise ValueError("Unknown differentiation scheme: {}".format(diff_scheme))

        if not isinstance(diff_type, DiffType):
            raise ValueError("Unknown differentiation type: {}".format(diff_type))

        if globalization_method is not None and not isinstance(
            globalization_method, GlobalMethod
        ):
            raise ValueError(
                "Unknown globalization method: {}".format(globalization_method)
            )

    def __check_evaluation_cost_type(
        self,
        evaluation_cost_type,  # type: Mapping[str, CostType]
    ):  # type: (...) -> None
        """Check the evaluation cost types.

        Args:
            evaluation_cost_type: The evaluation cost type of each function of the
                problem.

        Raises:
            ValueError: If a function name does not refer to a function of the problem,
                or if a cost type is invalid.
        """
        functions_names = [
            self.problem.get_objective_name()
        ] + self.problem.get_constraints_names()

        if evaluation_cost_type is not None:
            for func_name, func_type in evaluation_cost_type.items():
                if func_name not in functions_names:
                    raise ValueError("Unknown function name: {}".format(func_name))
                if not isinstance(func_type, CostType):
                    raise ValueError(
                        "Unknown cost type for function '{}': {}".format(
                            func_name, func_type
                        )
                    )

    def __check_expensive_evaluations(
        self,
        expensive_evaluations,  # type: Mapping[str, int]
    ):  # type: (...) -> None
        """Check the numbers of expensive evaluations.

        Args:
            expensive_evaluations: The maximal number of expensive evaluations for
                each function of the problem.

        Raises:
            ValueError: If a function name does not refer to a function of the problem.
            TypeError: If a number of expensive evaluations is not an integer.
        """
        functions_names = [
            self.problem.get_objective_name()
        ] + self.problem.get_constraints_names()

        for func_name, eval_number in expensive_evaluations.items():
            if func_name not in functions_names:
                raise ValueError("Unknown function name: {}".format(func_name))
            if not isinstance(eval_number, int):
                raise TypeError(
                    "Non-integer evaluations number for function '{}': {}".format(
                        func_name, eval_number
                    )
                )

    def __compute_constraints_tolerance(self):  # type: (...) -> Optional[float]
        """Compute the pSeven tolerance on the constraints.

        This tolerance is relative to the infinity norm.
        """
        tolerance = None

        if self.problem.has_eq_constraints() and self.problem.has_ineq_constraints():
            tolerance = min(self.problem.eq_tolerance, self.problem.ineq_tolerance)
        elif self.problem.has_eq_constraints():
            tolerance = self.problem.eq_tolerance
        elif self.problem.has_ineq_constraints():
            tolerance = self.problem.ineq_tolerance

        if tolerance:
            init_x = self.problem.design_space.get_current_x()
            constr_dim = sum(
                [constraint(init_x).size for constraint in self.problem.constraints]
            )
            tolerance /= sqrt(constr_dim)
            # N.B. ||c(x)||_2 <= sqrt(constr_dim) * ||c(x)||_inf

        return tolerance

    def __set_pseven_techniques(
        self,
        options,  # type: Dict[str, Any]
    ):  # type: (...) -> None
        """Get the pSeven techniques from the options."""
        techniques_list = list()
        internal_algo_name = self.lib_dict[self.algo_name][self.INTERNAL_NAME]

        if internal_algo_name in self.__LOCAL_METHODS:
            techniques_list.append(internal_algo_name)
        if options.get("globalization_method") is not None:
            techniques_list.append(options.pop("globalization_method"))
        if options.get("surrogate_based") is not None and options["surrogate_based"]:
            techniques_list.append(self.__SBO)
            options.pop("surrogate_based")

        options["GTOpt/Techniques"] = "[" + ", ".join(techniques_list) + "]"

    def _run(
        self, **options  # type: Any
    ):  # type: (...) -> OptimizationResult
        """Run the algorithm.

        Args:
            **options: The options of the algorithm.

        Returns:
            The result of the optimization.

        Raises:
            RuntimeError: If a solver for constrained problems is called on
                a problem without constraint.
            TypeError: If the QP solver is called on a problem
                whose objective is neither quadratic nor linear
                or one of whose constraints are not linear.
        """
        problem = self.problem
        options.pop("max_iter")

        # Check the functions
        if self.internal_algo_name in [
            self.__MOM,
            self.__QP,
            self.__SQP,
            self.__SQ2P,
        ]:
            if not problem.constraints:
                raise RuntimeError(
                    "{} requires at least one constraint".format(self.algo_name)
                )

        if self.internal_algo_name == self.__QP:
            if not isinstance(
                problem.nonproc_objective, (MDOQuadraticFunction, MDOLinearFunction)
            ):
                # TODO: support several objectives
                raise TypeError(
                    "{} requires the objective to be quadratic or linear".format(
                        self.algo_name
                    )
                )

            for constraint in problem.nonproc_constraints:
                if not isinstance(constraint, MDOLinearFunction):
                    raise TypeError(
                        "{} requires the constraints to be linear,"
                        " the following is not: {}".format(
                            self.algo_name, constraint.name
                        )
                    )

        # Grab the initial sample from the options
        sample = {
            "sample_x": options.pop("sample_x", []),
            "sample_f": options.pop("sample_f", []),
            "sample_c": options.pop("sample_c", []),
        }

        # Create the pSeven problem
        normalize_ds = options.pop(self.NORMALIZE_DESIGN_SPACE_OPTION, True)
        initial_x, lower_bnd, upper_bnd = self.get_x0_and_bounds_vects(normalize_ds)
        evaluation_cost_type = options.pop("evaluation_cost_type", None)
        expensive_evaluations = options.pop("expensive_evaluations", None)

        pseven_problem = PSevenProblem(
            problem,
            evaluation_cost_type,
            expensive_evaluations,
            lower_bnd,
            upper_bnd,
            initial_x,
        )

        # Run the algorithm and return the result
        try:
            result = p7core.gtopt.Solver().solve(
                pseven_problem, options=options, **sample
            )
        except p7core.exceptions.UserEvaluateException:
            # Gemseo terminated pSeven during evaluation as a stopping criterion was met
            status = p7core.status.USER_TERMINATED
        else:
            status = result.status

        return self.get_optimum_from_database(str(status), status.id)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Copyright ONERA, taken from WhatsOpt-CLI https://github.com/OneraHub/WhatsOpt-
CLI/blob/master/whatsopt/show_utils.py.

Distributed under the Apache 2.0 license

Minor modifications by Francois Gallard : merge the two methods an comment
"""
from typing import TYPE_CHECKING, Union

from gemseo.utils.py23_compat import Path

if TYPE_CHECKING:
    from gemseo.utils.xdsmizer import XdsmType

from xdsmjs import bundlejs, css

HTML_TEMPLATE = """
<!doctype html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
<style type="text/css">
{}
</style>
<script type="text/javascript">
{}
</script>
<script type="text/javascript">
    document.addEventListener('DOMContentLoaded', () => {{
      const mdo = {};
      const config = {{
        labelizer: {{
            ellipsis: 5,
            subSupScript: false,
            showLinkNbOnly: false,
        }},
        layout: {{
            origin: {{ x: 50, y: 20 }},
            cellsize: {{ w: 150, h: 50 }},
            padding: 10,
        }},
        withDefaultDriver: true,
        withTitleTooltip: true,
      }};
      xdsmjs.XDSMjs(config).createXdsm(mdo);
    }});
</script>
</head>
<body>
    <div class="xdsm-toolbar"></div>
    <div class="xdsm2"></div>
</body>
</html>
"""


def generate_xdsm_html(
    xdsm,  # type: XdsmType
    file_path="xdsm.html",  # type: Union[str,Path]
):  # type: (...) -> None
    """Generate a HTML file to visualize a dynamic and interactive XDSM.

    Args:
        xdsm: The XDSM structure.
        file_path: The name of the path to the output HTML file.
    """
    with open(str(file_path), "w") as stream:
        stream.write(HTML_TEMPLATE.format(css(), bundlejs(), xdsm))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Pierre-Jean Barjhoux
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A constraints plot."""
from __future__ import division, unicode_literals

import logging
from typing import Optional, Sequence, Tuple

import matplotlib.gridspec as gridspec
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import SymLogNorm
from matplotlib.ticker import MaxNLocator
from numpy import ndarray

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.post.core.colormaps import PARULA, RG_SEISMIC
from gemseo.post.opt_post_processor import OptPostProcessor
from gemseo.utils.py23_compat import PY2

LOGGER = logging.getLogger(__name__)


class ObjConstrHist(OptPostProcessor):
    """The constraint function history in line charts.

    By default, all the constraints are considered. A sublist of constraints can be
    passed as options.
    """

    def __init__(
        self,
        opt_problem,  # type: OptimizationProblem
    ):  # type: (...) -> None
        """
        Args:
            opt_problem: The optimization problem to be post-processed.
        """
        super(ObjConstrHist, self).__init__(opt_problem)
        self.opt_problem = opt_problem
        self.cmap = PARULA  # "viridis"  # "jet"
        self.ineq_cstr_cmap = RG_SEISMIC  # "seismic" "PRGn_r"
        self.eq_cstr_cmap = "seismic"  # "seismic" "PRGn_r"

    def _plot(
        self,
        constr_names=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> None
        """Create the design variables plot.

        Args:
            constr_names: The names of the constraints to plot.
                If None, use all the constraints.
        """
        obj_name = self.opt_problem.get_objective_name()
        obj_history, x_history, n_iter = self.__get_history(obj_name)
        fmin = np.min(obj_history)
        fmax = np.max(obj_history)
        if not self.opt_problem.minimize_objective and fmax < 0.0:
            obj_history = -obj_history
            fmin = np.min(obj_history)
            fmax = np.max(obj_history)
        grid = gridspec.GridSpec(1, 2, width_ratios=[15, 1], wspace=0.04, hspace=0.6)
        fig = plt.figure(figsize=(11, 6))
        ax1 = fig.add_subplot(grid[0, 0])
        # objective function
        plt.xlabel("Iterations", fontsize=12)
        plt.ylabel("Objective value", fontsize=12)
        plt.plot(np.arange(len(obj_history)), obj_history)
        plt.ylim([fmin, fmax])
        plt.xlim([0, n_iter])
        ax1.xaxis.set_major_locator(MaxNLocator(integer=True))
        plt.grid(True)
        plt.title("Evolution of the objective value and maximal constraint ")

        # Set window size
        mng = plt.get_current_fig_manager()
        mng.resize(700, 1000)

        # number of iterations
        nb_iter = x_history.shape[0]

        ineq_vals, eq_vals, ineq_cstr_id, eq_cstr_id = self.__get_constraints(
            constr_names=constr_names
        )

        # concatenate all constraints values (ineq + eq)
        # NB : we take absolute values of equality constraints for color map
        eq_vals_abs = np.abs(eq_vals)
        all_cstr_list_abs = [ineq_vals, eq_vals_abs]
        all_cstr_vals_abs = np.concatenate(
            [cstr for cstr in all_cstr_list_abs if cstr.size > 0], axis=1
        )
        maxi_by_iter = np.amax(all_cstr_vals_abs, axis=1)
        indmaxi_by_iter = np.argmax(all_cstr_vals_abs, axis=1)
        cstr_ids = np.concatenate((ineq_cstr_id, eq_cstr_id))

        values = np.atleast_2d(maxi_by_iter)
        cmap = RG_SEISMIC
        vmax = fmax
        vmin = -vmax
        extent = -0.5, nb_iter - 0.5, fmin, fmax

        # On python 2, base is not defined as a parameter in SymLogNorm()
        if PY2:
            norm = SymLogNorm(linthresh=1.0, vmin=vmin * 0.75, vmax=vmax * 0.75)
        else:
            norm = SymLogNorm(
                linthresh=1.0, vmin=vmin * 0.75, vmax=vmax * 0.75, base=np.e
            )

        im1 = ax1.imshow(
            values,
            cmap=cmap,
            interpolation="nearest",
            aspect="auto",
            extent=extent,
            norm=norm,
            alpha=0.6,
        )

        # add labels with constraint violation information
        all_cstr_list = [ineq_vals, eq_vals]
        all_cstr_vals = np.concatenate(
            [cstr for cstr in all_cstr_list if cstr.size > 0], axis=1
        )
        for iteration, ind in zip(range(nb_iter), indmaxi_by_iter):
            y_vals = 0.5 * (fmax + fmin)
            x_vals = iteration
            max_label = cstr_ids[ind]
            # no abs on equality constraint
            max_value = all_cstr_vals[iteration, ind]
            text = "constraint " + max_label + " = " + str(max_value)
            ax1.text(x_vals - 0.0, y_vals + 1.1 * fmin, text, rotation="vertical")

        # color map
        cax = fig.add_subplot(grid[0, 1])
        thick_min = int(np.log10(1.0))
        thick_max = int(np.log10(np.abs(vmax)))
        thick_num = thick_max - thick_min + 1
        levels_pos = np.logspace(thick_min, thick_max, num=thick_num, base=10.0)
        levels_pos = np.append(levels_pos, vmax)
        levels_neg = np.sort(-levels_pos)
        levels_neg = np.append(levels_neg, 0)
        levels = np.concatenate((levels_neg, levels_pos))

        col_bar = fig.colorbar(im1, cax=cax, ticks=levels)
        col_bar.ax.tick_params(labelsize=9)
        cax.set_xlabel("symlog")

        self._add_figure(fig)

    def __get_history(
        self,
        function_name,  # type: str
    ):  # type: (...) -> Tuple[ndarray,ndarray,int]
        """Access the optimization history of a function.

        Also the design variables at which it was computed.

        Args:
            function_name: The name of the function.

        Returns:
            The function values,
            the design variables values
            and the number of iterations.
        """
        f_hist, x_hist = self.database.get_func_history(function_name, x_hist=True)
        f_hist = np.array(f_hist).real
        x_hist = np.array(x_hist).real
        n_iter = x_hist.shape[0]

        return f_hist, x_hist, n_iter

    def __get_constraints(
        self,
        constr_names=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> Tuple[ndarray,ndarray,ndarray,ndarray]
        """Return the constraints with formatted shape.

        Args:
            constr_names: The names of the constraints.
                If None, use all the constraints.
        """
        # retrieve the constraints values
        ineq_cstr_names = []
        eq_cstr_names = []
        for cstr in self.opt_problem.get_ineq_constraints():
            if constr_names is None:
                ineq_cstr_names.append(cstr.name)
            else:
                if cstr.name in constr_names:
                    ineq_cstr_names.append(cstr.name)
        for cstr in self.opt_problem.get_eq_constraints():
            if constr_names is None:
                eq_cstr_names.append(cstr.name)
            else:
                if cstr.name in constr_names:
                    eq_cstr_names.append(cstr.name)
        get_hist_array = self.database.get_history_array
        if ineq_cstr_names:
            ineq_vals, ineq_id, _ = get_hist_array(ineq_cstr_names, add_dv=False)
        else:
            ineq_vals, ineq_id = np.array([]), np.array([])
        if eq_cstr_names:
            eq_vals, eq_cstr_id, _ = get_hist_array(eq_cstr_names, add_dv=False)
        else:
            eq_vals, eq_cstr_id = np.array([]), np.array([])

        # harmonization of tables format because constraints can be vectorial
        # or scalars. *vals.shape[0] = iteration, *vals.shape[1] = cstr values
        ineq_vals = np.atleast_3d(ineq_vals)
        ineq_shape = ineq_vals.shape
        ineq_vals = np.reshape(
            ineq_vals, (ineq_shape[0], ineq_shape[1] * ineq_shape[2])
        )
        eq_vals = np.atleast_3d(eq_vals)
        eq_shape = eq_vals.shape
        eq_vals = np.reshape(eq_vals, (eq_shape[0], eq_shape[1] * eq_shape[2]))

        return ineq_vals, eq_vals, ineq_id, eq_cstr_id



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

r"""Class to apply the OAT technique used by :class:`.MorrisIndices`.

OAT technique
-------------

The purpose of the One-At-a-Time (OAT) methodology is to quantify the elementary effect

.. math::

   df_i = f(X_1+dX_1,\ldots,X_i+dX_i,\ldots,X_d)-f(X_1,\ldots,X_i,\ldots,X_d)

associated with a small variation :math:`dX_i` of :math:`X_i` with

.. math::

   df_1 = f(X_1+dX_1,\ldots,X_i,\ldots,X_d)-f(X_1,\ldots,X_i,\ldots,X_d)

The elementary effects :math:`df_1,\ldots,df_d` are computed sequentially
from an initial point

.. math::

   X=(X_1,\ldots,X_d)

From these elementary effects, we can compare their absolute values
:math:`|df_1|,\ldots,|df_d|` and sort :math:`X_1,\ldots,X_d` accordingly.
"""

from __future__ import division, unicode_literals

import logging
from copy import deepcopy
from typing import Dict, Mapping, Tuple

from numpy import ndarray

from gemseo.algos.design_space import DesignSpace
from gemseo.core.discipline import MDODiscipline

LOGGER = logging.getLogger(__name__)


class OATSensitivity(MDODiscipline):
    """A :class:`.MDODiscipline` computing finite differences of another one.

    Args:
        discipline (MDODiscipline): A discipline.
        parameter_space (DesignSpace): A parameter space.
        step (float): A relative finite difference step between 0 and 0.5.
    """

    _PREFIX = "fd"

    def __init__(
        self,
        discipline,  # type: MDODiscipline
        parameter_space,  # type: DesignSpace
        step,  # type: float
    ):  # type: (...) -> None # noqa: D107
        super(OATSensitivity, self).__init__()
        inputs = parameter_space.variables_names
        self.input_grammar.initialize_from_data_names(inputs)
        outputs = [
            self.get_fd_name(input_, output)
            for output in discipline.get_output_data_names()
            for input_ in inputs
        ]
        self.output_grammar.initialize_from_data_names(outputs)
        self.discipline = discipline
        self.step = step
        self.parameter_space = parameter_space

    def _run(self):  # type: (...) -> None
        """Run method."""
        inputs = self.get_input_data()
        self.discipline.execute(inputs)
        out_prev = self.discipline.local_data
        for input_name in list(self.get_input_data_names()):
            inputs = self.__update_inputs(inputs, input_name, self.step)
            self.discipline.execute(inputs)
            out_curr = self.discipline.local_data
            out_diff = {
                name: out_curr[name] - out_prev[name]
                for name in self.discipline.get_output_data_names()
            }
            for name, value in out_diff.items():
                self.local_data[self.get_fd_name(input_name, name)] = value
            out_prev = out_curr

    @staticmethod
    def get_io_names(
        fd_name,  # type: str
    ):  # type: (...) -> Tuple[str,str]
        """Get the output and input names from finite difference name.

        Args:
            fd_name: A finite difference name.

        Returns:
            The output name.
            The input name
        """
        split_name = fd_name.split("!")
        output_name = split_name[1]
        input_name = split_name[2]
        return output_name, input_name

    def get_fd_name(
        self,
        input_name,  # type: str
        output_name,  # type: str
    ):  # type: (...) -> str
        """Return the output name associated to an input name.

        Args:
            input_name: An input name.
            output_name: An output name.

        Returns:
            The finite difference name.
        """
        return "{}!{}!{}".format(self._PREFIX, output_name, input_name)

    def __update_inputs(
        self,
        inputs,  # type: Mapping[str,ndarray]
        input_name,  # type:str
        step,  # type:float
    ):  # type: (...) -> Dict[str,ndarray]
        """Update the input data from a finite difference step and an input name.

        Args:
            inputs: The original input data.
            input_name: An input name.
            step: A relative finite difference step between 0 and 0.5.

        Returns:
            The updated input data.
        """
        if not 0 < step < 0.5:
            raise ValueError(
                "Relative finite difference step must be "
                "strictly comprised between 0 and 0.5; got {}".format(step)
            )
        inputs = deepcopy(inputs)
        l_b = self.parameter_space.get_lower_bound(input_name)
        u_b = self.parameter_space.get_upper_bound(input_name)
        abs_step = step * (u_b - l_b)
        if inputs[input_name] + abs_step > u_b:
            inputs[input_name] -= abs_step
        else:
            inputs[input_name] += abs_step
        return inputs



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""The polynomial model for regression.

Polynomial regression class is a particular case of the linear regression,
where the input data is transformed before the regression is applied.
This transform consists of creating a matrix of monomials (Vandermonde)
by raising the input data to different powers up to a certain degree :math:`D`.
In the case where there is only one input variable,
the input data :math:`(x_i)_{i=1, \dots, n}\in\mathbb{R}^n` is transformed
into the Vandermonde matrix

.. math::

    \begin{pmatrix}
        x_1^1  & x_1^2  & \cdots & x_1^D\\
        x_2^1  & x_2^2  & \cdots & x_2^D\\
        \vdots & \vdots & \ddots & \vdots\\
        x_n^1  & x_n^2  & \cdots & x_n^D\\
    \end{pmatrix}
    = (x_i^d)_{i=1, \dots, n;\ d=1, \dots, D}.

The output is expressed as a weighted sum of monomials:

.. math::

     y = w_0 + w_1 x^1 + w_2 x^2 + ... + w_D x^D,

where the coefficients :math:`(w_1, w_2, ..., w_d)` and the intercept :math:`w_0`
are estimated by least square regression.

In the case of a multidimensional input,
i.e. :math:`X = (x_{ij})_{i=1,\dots,n; j=1,\dots,m}`,
where :math:`n` is the number of samples and :math:`m` is the number of input variables,
the Vandermonde matrix is expressed
through different combinations of monomials of degree :math:`d, (1 \leq d \leq D)`;
e.g. for three variables :math:`(x, y, z)` and degree :math:`D=3`,
the different terms are
:math:`x`, :math:`y`, :math:`z`, :math:`x^2`, :math:`xy`, :math:`xz`,
:math:`y^2`, :math:`yz`, :math:`z^2`, :math:`x^3`, :math:`x^2y` etc.
More generally,
for m input variables,
the total number of monomials of degree :math:`1 \leq d \leq D` is given
by :math:`P = \binom{m+D}{m} = \frac{(m+D)!}{m!D!}`.
In the case of 3 input variables given above,
the total number of monomial combinations of degree lesser than or equal to three
is thus :math:`P = \binom{6}{3} = 20`.
The linear regression has to identify the coefficients :math:`(w_1, \dots, w_P)`,
in addition to the intercept :math:`w_0`.

This concept is implemented through the :class:`.PolynomialRegression` class
which inherits from the :class:`.MLRegressionAlgo` class.

Dependence
----------
The polynomial regression model relies on the LinearRegression class
of the `LinearRegression <https://scikit-learn.org/stable/modules/
linear_model.html>`_ and  `PolynomialFeatures <https://scikit-learn.org/stable/
modules/generated/sklearn.preprocessing.PolynomialFeatures.html>`_ classes of
the `scikit-learn library <https://scikit-learn.org/stable/modules/
linear_model.html>`_.
"""
from __future__ import division, unicode_literals

import logging
import pickle
from os.path import join
from typing import Iterable, Optional, Union

from numpy import concatenate, ndarray, where, zeros
from sklearn.preprocessing import PolynomialFeatures

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import DataType, TransformerType
from gemseo.mlearning.regression.linreg import LinearRegression

LOGGER = logging.getLogger(__name__)


class PolynomialRegression(LinearRegression):
    """Polynomial regression."""

    LIBRARY = "scikit-learn"
    ABBR = "PolyReg"

    def __init__(
        self,
        data,  # type: Dataset
        degree,  # type: int
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        fit_intercept=True,  # type: bool
        penalty_level=0.0,  # type: float
        l2_penalty_ratio=1.0,  # type: float
        **parameters  # type: Optional[Union[float,int,str,bool]]
    ):  # type: (...) -> None
        """
        Args:
            degree: The polynomial degree.
            fit_intercept: If True, fit intercept.
            penalty_level: The penalty level greater or equal to 0.
                If 0, there is no penalty.
            l2_penalty_ratio: The penalty ratio
                related to the l2 regularization.
                If 1, the penalty is the Ridge penalty.
                If 0, this is the Lasso penalty.
                Between 0 and 1, the penalty is the ElasticNet penalty.

        Raises:
            ValueError: If the degree is lower than one.
        """
        super(PolynomialRegression, self).__init__(
            data,
            degree=degree,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            fit_intercept=fit_intercept,
            penalty_level=penalty_level,
            l2_penalty_ratio=l2_penalty_ratio,
            **parameters
        )
        self._poly = PolynomialFeatures(degree=degree, include_bias=False)
        self.parameters["degree"] = degree
        if degree < 1:
            raise ValueError("Degree must be >= 1.")

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> None
        input_data = self._poly.fit_transform(input_data)
        super(PolynomialRegression, self)._fit(input_data, output_data)

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        input_data = self._poly.transform(input_data)
        return super(PolynomialRegression, self)._predict(input_data)

    def _predict_jacobian(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        # Dimensions:
        # powers:        (           ,            ,  n_powers ,  n_inputs )
        # coefs:         (           ,  n_outputs ,  n_powers ,           )
        # jac_coefs:     (           ,  n_outputs ,  n_powers ,  n_inputs )
        # vandermonde:   ( n_samples ,            ,  n_powers ,           )
        # contributions: ( n_samples ,  n_outputs ,  n_powers ,  n_inputs )
        # jacobians:     ( n_samples ,  n_outputs ,           ,  n_inputs )
        #
        # n_powers is given by the formula
        # n_powers = binom(n_inputs+degree, n_inputs)+1

        vandermonde = self._poly.transform(input_data)

        powers = self._poly.powers_
        n_inputs = self._poly.n_input_features_
        n_powers = self._poly.n_output_features_
        n_outputs = self.algo.coef_.shape[0]
        coefs = self.get_coefficients(False)

        jac_intercept = zeros((n_outputs, n_inputs))
        jac_coefs = zeros((n_outputs, n_powers, n_inputs))

        # Compute partial derivatives with respect to each input separately
        for index in range(n_inputs):

            # Coefficients of monomial derivatives
            dcoefs = powers[None, :, index] * coefs

            # Powers of monomial derivatives
            dpowers = powers.copy()
            dpowers[:, index] -= 1

            # Keep indices of remaining monomials only
            mask_zero = (dpowers == 0).prod(axis=1) == 1
            mask_keep = dpowers[:, index] >= 0
            mask_keep[mask_zero == 1] = False

            # Extract intercept for Jacobian (0th order term)
            dintercept = dcoefs[:, mask_zero].flatten()

            # Filter kept terms
            dcoefs = dcoefs[:, mask_keep]  # Coefficients of kept terms
            dpowers = dpowers[mask_keep]  # Power keys of kept terms

            # Find indices for the given powers
            inds_keep = [
                where((powers == dpowers[i]).prod(axis=1) == 1)
                for i in range(dpowers.shape[0])
            ]
            if len(inds_keep) > 0:
                inds_keep = concatenate(inds_keep).flatten()

            # Coefficients of partial derivatives in terms of original powers
            jac_intercept[:, index] = dintercept
            jac_coefs[:, inds_keep, index] = dcoefs

        # Assemble polynomial (sum of weighted monomials)
        contributions = jac_coefs[None] * vandermonde[:, None, :, None]
        jacobians = jac_intercept + contributions.sum(axis=2)

        return jacobians

    def get_coefficients(
        self,
        as_dict=False,  # type:bool
    ):  # type: (...) -> DataType
        """Return the regression coefficients of the linear model.

        Args:
            as_dict: If True, return the coefficients as a dictionary of Numpy arrays
                indexed by the names of the coefficients.
                Otherwise, return the coefficients as a Numpy array.
                For now the only valid value is False.

        Returns:
            The regression coefficients of the linear model.

        Raises:
            NotImplementedError: If the coefficients are required as a dictionary.
        """
        coefficients = self.coefficients
        if as_dict:
            raise NotImplementedError(
                "For now the coefficients can only be obtained "
                "in the form of a NumPy array"
            )
        return coefficients

    def _save_algo(
        self,
        directory,  # type: str
    ):  # type: (...) -> None
        super(PolynomialRegression, self)._save_algo(directory)
        filename = join(directory, "poly.pkl")
        with open(filename, "wb") as handle:
            pickle.dump(self._poly, handle)

    def load_algo(
        self,
        directory,  # type: str
    ):  # type: (...) -> None
        super(PolynomialRegression, self).load_algo(directory)
        filename = join(directory, "poly.pkl")
        with open(filename, "rb") as handle:
            poly = pickle.load(handle)
        self._poly = poly



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Iris dataset
============

This is one of the best known :class:`.Dataset`
to be found in the machine learning literature.

It was introduced by the statistician Ronald Fisher
in his 1936 paper "The use of multiple measurements in taxonomic problems",
Annals of Eugenics. 7 (2): 179–188.

It contains 150 instances of iris plants:

- 50 Iris Setosa,
- 50 Iris Versicolour,
- 50 Iris Virginica.

Each instance is characterized by:

- its sepal length in cm,
- its sepal width in cm,
- its petal length in cm,
- its petal width in cm.

This :class:`.Dataset` can be used for either clustering purposes
or classification ones.

`More information about the Iris dataset
<https://en.wikipedia.org/wiki/Iris_flower_data_set>`_

"""
from __future__ import division, unicode_literals

from os.path import abspath, dirname, join

from gemseo.core.dataset import Dataset


class IrisDataset(Dataset):
    """Iris dataset parametrization."""

    def __init__(self, name="Iris", by_group=True, as_io=False):
        """Constructor."""
        super(IrisDataset, self).__init__(name, by_group)
        dirpath = dirname(abspath(__file__))
        filename = join(dirpath, "iris.data")
        variables = [
            "sepal_length",
            "sepal_width",
            "petal_length",
            "petal_width",
            "specy",
        ]
        sizes = {
            "sepal_length": 1,
            "sepal_width": 1,
            "petal_length": 1,
            "petal_width": 1,
            "specy": 1,
        }
        if as_io:
            groups = {
                "sepal_length": "inputs",
                "sepal_width": "inputs",
                "petal_length": "inputs",
                "petal_width": "inputs",
                "specy": "outputs",
            }
        else:
            groups = {"specy": "labels"}
        self.set_from_file(filename, variables, sizes, groups, ",", False)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
OpenTUNRS DOE algorithms wrapper
********************************
"""
from __future__ import division, unicode_literals

import logging

import openturns
from matplotlib import pyplot as plt
from numpy import array
from numpy import max as np_max
from numpy import min as np_min
from numpy import ndarray
from openturns.viewer import View

from gemseo.algos.doe.doe_lib import DOELibrary

LOGGER = logging.getLogger(__name__)


class OpenTURNS(DOELibrary):
    """OpenTURNS library of DOE algorithms wrapping."""

    OT_DOC = "http://openturns.github.io/openturns/master/user_manual/"
    # Available algorithm for DOE design
    OT_SOBOL = "OT_SOBOL"
    OT_SOBOL_DESC = "Sobol sequence implemented in openTURNS library"
    OT_SOBOL_WEB = OT_DOC + "_generated/openturns.SobolSequence.html"
    OT_RANDOM = "OT_RANDOM"
    OT_RANDOM_DESC = "Random sampling implemented in openTURNS library"
    OT_RANDOM_WEB = OT_DOC + "_generated/openturns.RandomGenerator.html"
    OT_HASEL = "OT_HASELGROVE"
    OT_HASEL_DESC = "Haselgrove sequence implemented in openTURNS library"
    OT_HASEL_WEB = OT_DOC + "_generated/openturns.HaselgroveSequence.html"
    OT_REVERSE_HALTON = "OT_REVERSE_HALTON"
    OT_REVERSE_HALTON_DESC = (
        "Reverse Halton sequence implemented" " in openTURNS library"
    )
    OT_REVERSE_HALTON_WEB = OT_DOC + "_generated/openturns.ReverseHaltonSequence.html"
    OT_HALTON = "OT_HALTON"
    OT_HALTON_DESC = "Halton sequence implemented in openTURNS library"
    OT_HALTON_WEB = OT_DOC + "_generated/openturns.HaltonSequence.html"
    OT_FAURE = "OT_FAURE"
    OT_FAURE_DESC = "Faure sequence implemented in openTURNS library"
    OT_FAURE_WEB = OT_DOC + "_generated/openturns.FaureSequence.html"
    OT_MC = "OT_MONTE_CARLO"
    OT_MC_DESC = "Monte Carlo sequence implemented in openTURNS library"
    OT_MC_WEB = OT_DOC + "_generated/openturns.RandomGenerator.html"
    OT_FACTORIAL = "OT_FACTORIAL"
    OT_FACTORIAL_DESC = "Factorial design implemented in openTURNS library"
    OT_FACTORIAL_WEB = OT_DOC + "_generated/openturns.Factorial.html"
    OT_COMPOSITE = "OT_COMPOSITE"
    OT_COMPOSITE_DESC = "Composite design implemented in openTURNS library"
    OT_COMPOSITE_WEB = OT_DOC + "_generated/openturns.Composite.html"
    OT_AXIAL = "OT_AXIAL"
    OT_AXIAL_DESC = "Axial design implemented in openTURNS library"
    OT_AXIAL_WEB = OT_DOC + "_generated/openturns.Axial.html"
    OT_LHSO = "OT_OPT_LHS"
    OT_LHSO_DESC = "Optimal Latin Hypercube Sampling implemented in openTURNS library"
    OT_LHSO_WEB = (
        "https://openturns.github.io/openturns/master/examples/optimal_lhs.html"
    )
    OT_LHS = "OT_LHS"
    OT_LHS_DESC = "Latin Hypercube Sampling implemented in openTURNS library"
    OT_LHS_WEB = OT_DOC + "_generated/openturns.LHS.html"
    OT_LHSC = "OT_LHSC"
    OT_LHSC_DESC = (
        "Centered Latin Hypercube Sampling implemented" " in openTURNS library"
    )
    OT_LHSC_WEB = OT_DOC + "_generated/openturns.LHS.html"
    OT_FULLFACT = "OT_FULLFACT"  # Box in openturns
    OT_FULLFACT_DESC = "Full factorial design implemented" "in openTURNS library"
    OT_FULLFACT_WEB = OT_DOC + "_generated/openturns.Box.html"
    OT_SOBOL_INDICES = "OT_SOBOL_INDICES"
    OT_SOBOL_INDICES_DESC = "Sobol indices"
    OT_SOBOL_INDICES_WEB = OT_DOC + "_generated/openturns.SobolIndicesAlgorithm.html"

    ALGO_LIST = [
        OT_SOBOL,
        OT_HASEL,
        OT_REVERSE_HALTON,
        OT_HALTON,
        OT_FAURE,
        OT_AXIAL,
        OT_FACTORIAL,
        OT_MC,
        OT_LHS,
        OT_LHSC,
        OT_LHSO,
        OT_RANDOM,
        OT_FULLFACT,
        OT_COMPOSITE,
        OT_SOBOL_INDICES,
    ]
    DESC_LIST = [
        OT_SOBOL_DESC,
        OT_HASEL_DESC,
        OT_REVERSE_HALTON_DESC,
        OT_HALTON_DESC,
        OT_FAURE_DESC,
        OT_AXIAL_DESC,
        OT_FACTORIAL_DESC,
        OT_MC_DESC,
        OT_LHS_DESC,
        OT_LHSC_DESC,
        OT_LHSO_DESC,
        OT_RANDOM_DESC,
        OT_FULLFACT_DESC,
        OT_COMPOSITE_DESC,
        OT_SOBOL_INDICES_DESC,
    ]
    WEB_LIST = [
        OT_SOBOL_WEB,
        OT_HASEL_WEB,
        OT_REVERSE_HALTON_WEB,
        OT_HALTON_WEB,
        OT_FAURE_WEB,
        OT_AXIAL_WEB,
        OT_FACTORIAL_WEB,
        OT_MC_WEB,
        OT_LHS_WEB,
        OT_LHSC_WEB,
        OT_LHSO_WEB,
        OT_RANDOM_WEB,
        OT_FULLFACT_WEB,
        OT_COMPOSITE_WEB,
        OT_SOBOL_INDICES_WEB,
    ]

    # Available distribution (only a part of what is available in openturns
    OT_ARCSINE = "Arcsine"
    OT_BETA = "Beta"
    OT_DIRICHLET = "Dirichlet"
    OT_NORMAL = "Normal"
    OT_TRUNCNORMAL = "TruncatedNormal"
    OT_TRIANGULAR = "Triangular"
    OT_TRAPEZOIDAL = "Trapezoidal"
    OT_UNIFORM = "Uniform"

    DISTRIBUTION_LIST = [
        OT_ARCSINE,
        OT_BETA,
        OT_DIRICHLET,
        OT_NORMAL,
        OT_TRUNCNORMAL,
        OT_TRIANGULAR,
        OT_TRAPEZOIDAL,
        OT_UNIFORM,
    ]

    # Optional parameters
    LEVEL_KEYWORD = "levels"
    CENTER_KEYWORD = "centers"
    DISTRIBUTION_KEYWORD = "distribution_name"
    MEAN_KEYWORD = "mu"
    STD_KEYWORD = "sigma"
    START_KEYWORD = "start"
    END_KEYWORD = "end"
    DOE_SETTINGS_OPTIONS = [
        LEVEL_KEYWORD,
        DISTRIBUTION_KEYWORD,
        MEAN_KEYWORD,
        STD_KEYWORD,
        START_KEYWORD,
        END_KEYWORD,
        CENTER_KEYWORD,
    ]
    CRITERION = "criterion"
    CRITERIA = {
        "C2": openturns.SpaceFillingC2,
        "PhiP": openturns.SpaceFillingPhiP,
        "MinDist": openturns.SpaceFillingMinDist,
    }
    TEMPERATURE = "temperature"
    TEMPERATURES = {
        "Geometric": openturns.GeometricProfile,
        "Linear": openturns.LinearProfile,
    }
    N_REPLICATES = "n_replicates"
    ANNEALING = "annealing"

    # Default parameters
    DISTRIBUTION_DEFAULT = OT_UNIFORM

    def __init__(self):
        """Constructor Unless mentioned, DOE are normalized between [0,1]"""
        super(OpenTURNS, self).__init__()
        self.__distr_list = []
        self.__comp_dist = None
        self.__sequence = None
        for idx, algo in enumerate(self.ALGO_LIST):
            self.lib_dict[algo] = {
                DOELibrary.LIB: self.__class__.__name__,
                DOELibrary.INTERNAL_NAME: algo,
                DOELibrary.DESCRIPTION: self.DESC_LIST[idx],
                DOELibrary.WEBSITE: self.WEB_LIST[idx],
            }

    def _get_options(
        self,
        distribution_name="Uniform",  # pylint: disable=W0221
        levels=None,
        centers=None,
        eval_jac=False,
        n_samples=1,
        mu=0.5,
        sigma=None,
        start=0.25,
        end=0.75,
        n_processes=1,
        wait_time_between_samples=0.0,
        criterion="C2",
        temperature="Geometric",
        annealing=True,
        n_replicates=1000,
        seed=1,
        max_time=0,
        **kwargs
    ):
        """Sets the options.

        :param distribution_name: distribution name
        :type distribution_name: str
        :param levels: levels for axial, factorial and composite designs
        :type levels: array
        :param centers: centers for axial, factorial and composite designs
        :type centers: array
        :param eval_jac: evaluate jacobian
        :type eval_jac: bool
        :param n_samples: number of samples
        :type n_samples: int
        :param mu: mean of a random variable for beta, normal and
            truncated normal distributions
        :type mu: float
        :param sigma: standard deviation for beta, normal and
            truncated normal distributions
        :type sigma: float
        :param start: level start for trapezoidal distribution
        :type start: float
        :param end: level end for trapezoidal distribution
        :type end: float
        :param n_processes: number of processes
        :type n_processes: int
        :param wait_time_between_samples: waiting time between two samples
        :type wait_time_between_samples: float
        :param criterion: space-filling criterion, either "C2", "PhiP" or "MinDist".
            Default: "C2".
        :type criterion: str
        :param temperature: temperature profil for simulated annealing,
            either "Geometric" or "Linear". Default: "Geometric".
        :param annealing: if True, use simulated annealing to optimize LHS. Otherwise,
            use crude Monte Carlo. Default: True.
        :type annealing: bool
        :param n_replicates: number of Monte Carlo replicates to optimize LHS.
            Default: 1000.
        :type n_replicates: int
        :param seed: seed value.
        :type seed: int
        :param max_time: maximum runtime in seconds,
            disabled if 0 (Default value = 0)
        :type max_time: float
        :param kwargs: additional arguments
        """
        if levels is None:
            levels = [0.0, 0.25, 0.5]
        if centers is None:
            centers = [0.5]
        if sigma is None:
            sigma = 0.447214 * 0.5
        wtbs = wait_time_between_samples
        popts = self._process_options(
            distribution_name=distribution_name,
            levels=levels,
            centers=centers,
            eval_jac=eval_jac,
            n_samples=n_samples,
            mu=mu,
            sigma=sigma,
            start=start,
            end=end,
            n_processes=n_processes,
            wait_time_between_samples=wtbs,
            criterion=criterion,
            temperature=temperature,
            annealing=annealing,
            n_replicates=n_replicates,
            seed=seed,
            max_time=max_time,
            **kwargs
        )

        return popts

    def __set_level_option(self, options):
        """Check that level options is properly defined for stratified DOE.

        :param options: the options dict for the DOE
        """
        option = options[self.LEVEL_KEYWORD]
        if isinstance(option, (list, tuple)):
            option = array(option)
            if np_max(option) > 1.0:
                raise ValueError(
                    "Upper bound of levels must be <=1: %s given" % np_max(option)
                )
            if np_min(option) < 0.0:
                raise ValueError(
                    "Lower bound of levels must be >=1: %s given" % np_min(option)
                )
            options[self.LEVEL_KEYWORD] = option
        else:
            raise TypeError(
                "Error for levels definition in DOE design:"
                + "a tuple or a list is expected whereas ",
                type(option),
                " is provided",
            )
        return options

    def __set_center_option(self, dimension, options):
        """Check that center level options is properly defined for stratified DOE.

        :param str dimension: parameter space dimension.
        :param options: the options dict for the DOE
        """
        center = options[self.CENTER_KEYWORD]
        if isinstance(center, (list, tuple)):
            if len(center) != dimension:
                raise ValueError(
                    "Inconsistent length of 'centers' list argument "
                    + "compared to design vector size: %s vs %s"
                    % (dimension, len(center))
                )
            options[self.CENTER_KEYWORD] = array(center)
        else:
            raise TypeError(
                "Error for 'centers' definition in DOE design:"
                + "a tuple or a list is expected whereas ",
                type(center),
                " is provided",
            )
        return options

    def __get_distribution(self, options):
        """If no distribution is provided (a name or a list of composed distributions)
        then a default setting is done.

        :param options: the options dict for the distribution
        """
        if self.DISTRIBUTION_KEYWORD in options:
            distribution_name = options[self.DISTRIBUTION_KEYWORD]
            del options[self.DISTRIBUTION_KEYWORD]
        else:
            distribution_name = self.DISTRIBUTION_DEFAULT
        return distribution_name, options

    def _generate_samples(self, **options):
        """Generates the list of x samples.

        :param options: the options dict for the algorithm,
            see associated JSON file
        """
        self.seed += 1
        dimension = options[self.DIMENSION]
        del options[self.DIMENSION]
        n_samples = options[self.N_SAMPLES]
        del options[self.N_SAMPLES]
        LOGGER.info("Generation of %s DOE with OpenTurns", self.algo_name)
        if self.algo_name in (self.OT_LHS, self.OT_LHSC, self.OT_LHSO):
            distribution_name, options = self.__get_distribution(options)
            samples = self.__generate_lhs(
                n_samples, dimension, distribution_name=distribution_name, **options
            )
        elif self.algo_name == self.OT_RANDOM:
            samples = self.__generate_random(n_samples, dimension, **options)
        elif self.algo_name == self.OT_MC:
            distribution_name, options = self.__get_distribution(options)
            samples = self.__generate_mc(
                n_samples, dimension, distribution_name=distribution_name, **options
            )
        elif self.algo_name == self.OT_FULLFACT:
            samples = self.__generate_fullfact(n_samples, dimension)
        elif self.algo_name in (self.OT_COMPOSITE, self.OT_AXIAL, self.OT_FACTORIAL):
            options = self.__check_stratified_options(dimension, options)
            samples = self.__generate_stratified(options)
        elif self.algo_name == self.OT_SOBOL_INDICES:
            samples = self.__generate_sobol(n_samples, dimension, **options)
        else:
            samples = self.__generate_seq(n_samples, dimension)
        return samples

    @staticmethod
    def __check_float(options, keyword, default=0, u_b=None, l_b=None):
        """Base function to check if the keyword exist in dictionary and set a default
        value.

        :param options: dictionary of optional parameters
        :type  options: dictionary
        :param keyword: name of the optional keyword
        :type  keyword: string
        :param default: default value
        :type  default: float
        :param u_b: upper bound
        :type  u_b: float
        :param l_b: lower bound
        :type  l_b: float
        """
        if keyword in options:
            opt = options[keyword]
            if not isinstance(opt, float):
                raise TypeError(
                    keyword + " value must be a float : ", type(opt), " given"
                )
            if u_b is not None:
                if opt > u_b:
                    raise ValueError(
                        keyword
                        + " value must be < "
                        + str(u_b)
                        + " : "
                        + str(opt)
                        + " given"
                    )
            if l_b is not None:
                if opt < l_b:
                    raise ValueError(
                        keyword
                        + " value must be > "
                        + str(l_b)
                        + " : "
                        + str(opt)
                        + " given"
                    )
            options[keyword] = opt
        else:
            options[keyword] = default
        return options

    def __check_stratified_options(self, dimension, options):
        """Check that mandatory inputs for composite design are set.

        :param int dimension: parameter space dimension.
        :param options: the options
        """
        if self.LEVEL_KEYWORD not in options:
            raise KeyError(
                "Missing  parameter 'levels', "
                + "tuple of normalized levels  "
                + "in [0,1] you need in your design"
            )
        options = self.__set_level_option(options)
        if self.CENTER_KEYWORD not in options:
            options[self.CENTER_KEYWORD] = [0.5 for _ in range(dimension)]
        else:
            options = self.__set_center_option(dimension, options)
        return options

    def __generate_stratified(self, options):
        """Generate a DOE using composite algo of openturns.

        :param options: the options
        :returns: samples
        :rtype: numpy array
        """
        levels = options[self.LEVEL_KEYWORD]
        centers = options[self.CENTER_KEYWORD]
        LOGGER.info("Composite design:")
        LOGGER.info("    - centers: %s", str(centers))
        LOGGER.info("    - levels: %s", str(levels))
        if self.algo_name == self.OT_COMPOSITE:
            experiment = openturns.Composite(centers, levels)
        elif self.algo_name == self.OT_AXIAL:
            experiment = openturns.Axial(centers, levels)
        elif self.algo_name == self.OT_FACTORIAL:
            experiment = openturns.Factorial(centers, levels)

        samples = array(experiment.generate())
        samples = self._rescale_samples(samples)
        return samples

    def __generate_seq(self, n_samples, dimension):
        """Generate a DOE using LHS algo of openturns.

        :param n_samples: number of samples in DOE
        :type  n_samples: integer
        :param int dimension: parameter space dimension.
        :returns: samples
        :rtype: numpy array
        """
        if self.algo_name == self.OT_FAURE:
            self.__sequence = openturns.FaureSequence
        elif self.algo_name == self.OT_HALTON:
            self.__sequence = openturns.HaltonSequence
        elif self.algo_name == self.OT_REVERSE_HALTON:
            self.__sequence = openturns.ReverseHaltonSequence
        elif self.algo_name == self.OT_HASEL:
            self.__sequence = openturns.HaselgroveSequence
        elif self.algo_name == self.OT_SOBOL:
            self.__sequence = openturns.SobolSequence
        seq = self.__sequence(dimension).generate(n_samples)
        return array(seq)

    def create_composed_distributions(self):
        """Create a composed distribution from a list of distributions."""
        self.__comp_dist = openturns.ComposedDistribution(self.__distr_list)

    def get_composed_distributions(self):
        """Returns the composed distributions.

        :returns: composed distributions
        :rtype: openturns.ComposedDistribution
        """
        return self.__comp_dist

    def __check_composed_distribution(self, distribution_name, dimension):
        """Checks the composed distribution.

        :param str distribution_name: name of the distribution
        :param int dimension: parameter space dimension.
        """
        if self.__comp_dist is None:
            n_distrib = len(self.__distr_list)
            if n_distrib == 0:
                LOGGER.info(
                    "Creating default composed distribution based on %s",
                    distribution_name,
                )
                self.create_distribution(distribution_name)
                self.__comp_dist = openturns.ComposedDistribution(
                    [self.__distr_list[0] for _ in range(dimension)]
                )
            elif n_distrib == 1:
                # Only one distribution was defined ==> duplicating it in all
                # dimensions
                self.__comp_dist = openturns.ComposedDistribution(
                    [self.__distr_list[0] for _ in range(dimension)]
                )
            elif n_distrib != dimension:
                raise ValueError(
                    "Size mismatch between number"
                    " of distribution and problem: "
                    "{} vs. {}".format(dimension, n_distrib)
                )
        elif self.__comp_dist.getDimension() != dimension:
            raise ValueError(
                "Size mismatch between ComposedDistribution and "
                "problem: {} vs. {}".format(dimension, self.__comp_dist.getDimension())
            )
        else:
            LOGGER.info(
                "Using composed distribution previously created: %s",
                str(self.__comp_dist.getDistributionCollection()),
            )

    def check_distribution_name(self, distribution_name):
        """Check that distribution is available.

        :param distribution_name: name of the distribution
        :type distribution_name: string
        """
        if distribution_name not in self.DISTRIBUTION_LIST:
            raise ValueError(
                "Distribution '"
                + distribution_name
                + "' is not available. "
                + "Please switch to one of the followings: "
                + str(self.DISTRIBUTION_LIST)
            )

    def create_distribution(self, distribution_name="Uniform", **options):
        """Create a distribution for all design vectors and add it to the list of
        distributions.

        :param distribution_name: name of the distribution
           (Default value = "Uniform")
        :type distribution_name: str
        :param options: optional parameters
        :type options: dict
        :param options: OT distributions options
        """
        self.check_distribution_name(distribution_name)
        options = self.__check_float(options, self.MEAN_KEYWORD, 0.5, u_b=1, l_b=0)
        options = self.__check_float(options, self.CENTER_KEYWORD, 0.5, u_b=1, l_b=0)
        options = self.__check_float(options, self.START_KEYWORD, 0.25, u_b=1, l_b=0)
        options = self.__check_float(options, self.END_KEYWORD, 0.75, u_b=1, l_b=0)

        if distribution_name == self.OT_UNIFORM:
            LOGGER.info("Creation of a uniform distribution")
            self.__distr_list.append(openturns.Uniform(0, 1))
        elif distribution_name == self.OT_TRIANGULAR:
            dist_center = options[self.CENTER_KEYWORD]
            LOGGER.info(
                "Creation of a triangular distribution" " with center: %s",
                str(dist_center),
            )
            self.__distr_list.append(openturns.Triangular(0.0, dist_center, 1.0))
        elif distribution_name == self.OT_TRAPEZOIDAL:
            lower_bound = options[self.START_KEYWORD]
            upper_bound = options[self.END_KEYWORD]
            LOGGER.info(
                "Creation of a trapezoidal distribution "
                "with lower/upper bounds: %s %s",
                str(lower_bound),
                str(upper_bound),
            )
            self.__distr_list.append(
                openturns.Trapezoidal(0.0, lower_bound, upper_bound, 1.0)
            )
        elif distribution_name == self.OT_BETA:
            mean = options[self.MEAN_KEYWORD]
            options = self.__check_float(
                options, self.STD_KEYWORD, 0.447214 * 0.5, u_b=1, l_b=0
            )
            std = options[self.STD_KEYWORD]
            LOGGER.info(
                "Creation of a %s distribution" ": mu (mean) %g and sigma (std) %g",
                distribution_name,
                mean,
                std,
            )
            beta = openturns.Beta()
            beta.setParameter(openturns.BetaMuSigma()([mean, std, 0.0, 1.0]))
            self.__distr_list.append(beta)
        elif distribution_name == self.OT_ARCSINE:
            LOGGER.info("Creation of a %s", str(distribution_name))
            arcs = openturns.Arcsine(0.0, 1.0)
            self.__distr_list.append(arcs)
        elif distribution_name == self.OT_TRUNCNORMAL:
            mean = options[self.MEAN_KEYWORD]
            options = self.__check_float(
                options, self.STD_KEYWORD, 0.5 / 3.75, u_b=1, l_b=0
            )
            std = options[self.STD_KEYWORD]
            LOGGER.info(
                "Creation of a %s distribution: mu (mean) %g and" "sigma (std) %g",
                distribution_name,
                mean,
                std,
            )
            nrmal = openturns.TruncatedNormal(mean, std, 0.0, 1.0)
            self.__distr_list.append(nrmal)
        elif distribution_name == self.OT_NORMAL:
            mean = options[self.MEAN_KEYWORD]
            options = self.__check_float(
                options, self.STD_KEYWORD, 0.5 / 3.75, u_b=1, l_b=0
            )
            std = options[self.STD_KEYWORD]
            LOGGER.info(
                "Creation of a %s distribution: mu (mean) %g and" "sigma (std) %g",
                distribution_name,
                mean,
                std,
            )
            self.__distr_list.append(openturns.Normal(mean, std))

    def display_distributions_list(self):
        """Display list of distributions use or that will be used for DOE design based
        on LHS or Monte-Carlo methods."""
        LOGGER.info("List of distributions:")
        for distrib in self.__distr_list:
            LOGGER.info(distrib)

    def get_distributions_list(self):
        """Accessor for distributions list.

        :returns: distribution list
        :rtype: list
        """
        return self.__distr_list

    def __generate_lhs(
        self, n_samples, dimension, distribution_name="Uniform", **options
    ):
        """Generate a DOE using LHS algo of openturns.

        :param int n_samples: number of samples in DOE
        :param int dimension: parameter space dimension.
        :param distribution_name: name of the distribution
        :type  distribution_name: string
        :param options: the options
        :returns: samples
        :rtype: numpy array
        """
        self.__check_composed_distribution(
            distribution_name=distribution_name, dimension=dimension
        )

        seed = options.get(self.SEED, self.seed)
        openturns.RandomGenerator.SetSeed(seed)
        lhs = openturns.LHSExperiment(self.__comp_dist, n_samples)
        if self.algo_name == self.OT_LHSO:
            lhs.setAlwaysShuffle(True)
            criterion = options.get(self.CRITERION, "C2")
            try:
                criterion = self.CRITERIA[criterion]()
            except KeyError:
                raise ValueError(
                    "{} is not an available criterion. Available ones are: {}".format(
                        criterion, self.CRITERIA
                    )
                )
            annealing = options.get(self.ANNEALING, True)
            if annealing:
                temperature = options.get(self.TEMPERATURE, "Geometric")
                try:
                    temperature = self.TEMPERATURES[temperature]()
                except KeyError:
                    raise ValueError(
                        "{} is not an available temperature profil."
                        "Available ones are: {}".format(temperature, self.TEMPERATURES)
                    )
                algo = openturns.SimulatedAnnealingLHS(lhs, temperature, criterion)
                design = algo.generate()
            else:
                n_replicates = options.get(self.N_REPLICATES, 1000)
                algo = openturns.MonteCarloLHS(lhs, n_replicates)
                design = algo.generate()
        else:
            design = lhs.generate()

        samples = array(design)
        if self.algo_name == self.OT_LHSC:
            samples = self.__compute_centered_lhs(samples)
        return samples

    @staticmethod
    def __compute_centered_lhs(
        samples,  # type:ndarray
    ):  # type:(...) ->ndarray
        """Center the samples resulting from a Latin hypercube sampling.

        Args:
            samples: The samples resulting from a Latin hypercube sampling.

        Returns:
            The centered version of the initial samples.
        """
        n_samples = len(samples)
        centered_samples = (samples // (1.0 / n_samples) + 0.5) / n_samples
        return centered_samples

    def __generate_mc(
        self, n_samples, dimension, distribution_name="Uniform", **options
    ):
        """Generate a DOE using Monte-Carlo algo of openturns.

        :param int n_samples: number of samples in DOE
        :param int dimension: parameter space dimension
        :param distribution_name: name of the distribution
        :type  distribution_name: string
        :param options: the options
        :returns: samples
        :rtype: numpy array
        """
        self.__check_composed_distribution(
            distribution_name=distribution_name, dimension=dimension
        )

        seed = options.get(self.SEED, self.seed)
        openturns.RandomGenerator.SetSeed(seed)
        experiment = openturns.MonteCarloExperiment(self.__comp_dist, n_samples)
        return array(experiment.generate())

    def __generate_sobol(self, n_samples, dimension, **options):
        """Generate a DOE using Sobol' sampling.

        :param int n_samples: number of samples in DOE
        :param int dimension: parameter space dimension
        :returns: samples
        :rtype: numpy array
        """
        seed = options.get(self.SEED, self.seed)
        openturns.RandomGenerator.SetSeed(seed)
        self.__check_composed_distribution(
            distribution_name="Uniform", dimension=dimension
        )
        n_samples = int(n_samples / (dimension + 2))
        experiment = openturns.SobolIndicesExperiment(self.__comp_dist, n_samples)
        data = array(experiment.generate())
        return data

    def __generate_fullfact(self, n_samples, dimension):
        """Generate a DOE using Monte-Carlo algo of openturns.

        :param int n_samples: number of samples in DOE
        :param int dimension: parameter space dimension.
        :returns: samples
        :rtype: numpy array
        """
        self._display_fullfact_warning(n_samples)
        level = int(n_samples ** (1.0 / dimension) - 2)
        if level < 1:
            level = 0
        levels = [level] * dimension
        experiment = openturns.Box(levels)
        return array(experiment.generate())

    def __generate_random(self, n_samples, dimension, **options):
        """Generate a DOE using random algo of openturns.

        :param n_samples: number of samples in DOE
        :type  n_samples: integer
        :param int dimension: parameter space dimension
        :returns: samples
        :rtype: numpy array
        """
        seed = options.get(self.SEED, self.seed)
        openturns.RandomGenerator.SetSeed(seed)
        samples_list = []
        for _ in range(n_samples):
            samples_list.append(openturns.RandomGenerator.Generate(dimension))
        return array(samples_list)

    @staticmethod
    def plot_distribution(distribution, show=False):
        """Plot the density PDF & the CDF (cumulative) of a given distribution.

        :param distribution: the distribution to plot
        :type distribution: openturns.Distribution
        :param show: show plot (Default value = False)
        :type show: bool
        """
        distribution.setDescription(["x"])
        pdf_graph = distribution.drawPDF()
        cdf_graph = distribution.drawCDF()
        fig = plt.figure(figsize=(12, 5))
        plt.suptitle(str(distribution))
        pdf_axis = fig.add_subplot(121)
        cdf_axis = fig.add_subplot(122)
        mean = array(distribution.getMean())[0]
        std = array(distribution.getStandardDeviation())[0]
        View(pdf_graph, figure=fig, axes=[pdf_axis], add_legend=False)
        pdf_axis.axvline(x=mean, ymin=0.0, linewidth=1, linestyle="-.", color="k")
        pdf_axis.axvline(
            x=mean - std, linestyle="--", ymin=0.0, linewidth=0.5, color="k"
        )
        pdf_axis.axvline(
            x=mean + std, linestyle="--", ymin=0.0, linewidth=0.5, color="k"
        )
        pdf_axis.annotate(
            r"$\mu$",
            xy=(0.5, 1.05),
            xycoords="axes fraction",
            horizontalalignment="center",
            verticalalignment="center",
            size=15,
        )
        pdf_axis.annotate(
            r"$\mu+\sigma$",
            xy=(mean + std, 1.05),
            xycoords="axes fraction",
            horizontalalignment="center",
            verticalalignment="center",
            size=15,
        )
        pdf_axis.annotate(
            r"$\mu-\sigma$",
            xy=(mean - std, 1.05),
            xycoords="axes fraction",
            horizontalalignment="center",
            verticalalignment="center",
            size=15,
        )
        View(cdf_graph, figure=fig, axes=[cdf_axis], add_legend=False)
        if show:
            plt.show()
        plt.close()
